from .abf import *
from .eabf import *
from .enhanced_sampling import *
from .gamd import *
from .gawtmeabf import *
from .metadynamics import *
from .metaeabf import *
from .reference import *
from .utils import *

__all__ = [
    "eabf",
    "eabf",
    "enhanced_sampling",
    "gamd",
    "gawtmeabf",
    "metadynamics",
    "metaeabf",
    "reference",
    "utils",
]



import sys
import torch
import numpy as np

from typing import Union, Tuple
from ..interface.sampling_data import MDInterface
from ..units import *
from .utils import *


class CV:
    """Class for Collective Variables

    Automatic computation of gradients is implemented using the autograd differentiation engine of Torch

    Args:
        the_mol: Molecule Object containing masses (natoms) and coords (3*natoms)
        requires_grad: if True, partial derivatives of CVs with respect
                                to atom coordinats are computed
                                and saved to self.gradient
    """

    def __init__(self, the_mol: MDInterface, requires_grad: bool = False):

        self.the_mol = the_mol
        self.requires_grad = requires_grad

        md_state = self.the_mol.get_sampling_data()

        self.mass = torch.from_numpy(md_state.mass).float()
        self.coords = torch.from_numpy(md_state.coords.ravel()).float()
        self.natoms = len(self.mass)
        self.requires_grad = requires_grad
        self.gradient = None
        self.cv = None
        self.type = "default"
        self.reference = None
        self.reference_internal = None

    def update_coords(self):
        """The coords tensor and ndarray share the same memory.
        Modifications to the tensor will be reflected in the ndarray and vice versa!"""
        self.coords = torch.from_numpy(self.the_mol.get_sampling_data().coords.ravel())
        self.coords = self.coords.float()
        self.coords.requires_grad = self.requires_grad

    def _get_com(self, atoms: Union[int, list]) -> Tuple[torch.tensor, float]:
        """get center of mass (com) of group of atoms

        Args:
            atoms (Union[int, list]): atom index or list of atom indices

        Returns:
            com (torch.tensor): Center of Mass
        """
        if hasattr(atoms, "__len__"):
            # compute center of mass for group of atoms
            center = torch.matmul(
                self.coords.view((self.natoms, 3))[atoms].T, self.mass[atoms]
            )
            m_tot = self.mass[atoms].sum()
            com = center / m_tot

        else:
            # only one atom
            atom = int(atoms)
            com = self.coords[3 * atom : 3 * atom + 3]

        return com

    def x(self) -> float:
        """use x axis as cv for numerical examples"""
        self.update_coords()
        self.cv = self.coords[0]
        if self.requires_grad:
            self.gradient = torch.autograd.grad(
                self.cv, self.coords, allow_unused=True
            )[0]
            self.gradient = self.gradient.detach().numpy()

        return float(self.cv)

    def y(self) -> float:
        """use y axis as cv for numerical examples"""
        self.update_coords()
        self.cv = self.coords[1]
        if self.requires_grad:
            self.gradient = torch.autograd.grad(
                self.cv, self.coords, allow_unused=True
            )[0]
            self.gradient = self.gradient.detach().numpy()

        return float(self.cv)

    def distance(self, cv_def: list) -> float:
        """distance between two mass centers in range(0, inf)

        Args:
            cv_def (list):
                distance beteen atoms: [ind0, ind1]
                distance between mass centers: [[ind00, ind01, ...], [ind10, ind11, ...]]

        Returns:
            cv (float): computed distance
        """
        if len(cv_def) != 2:
            raise ValueError(
                "CV ERROR: Invalid number of centers in definition of distance!"
            )

        self.update_coords()

        p1 = self._get_com(cv_def[0])
        p2 = self._get_com(cv_def[1])

        # get distance
        r12 = p2 - p1
        self.cv = torch.linalg.norm(r12)

        # get forces
        if self.requires_grad:
            self.gradient = torch.autograd.grad(
                self.cv, self.coords, allow_unused=True
            )[0]
            self.gradient = self.gradient.detach().numpy()

        return float(self.cv)

    def angle(self, cv_def: list) -> float:
        """get angle between three mass centers in range(-pi,pi)

        Args:
            cv_def (list):
                angle between two atoms: [ind0, ind1, ind3]
                angle between centers of mass: [[ind00, ind01, ...], [ind10, ind11, ...], [ind20, ind21, ...]]

        Returns:
            cv (float): computed angle
        """
        if len(cv_def) != 3:
            raise ValueError(
                "CV ERROR: Invalid number of centers in definition of angle!"
            )

        self.update_coords()

        p1 = self._get_com(cv_def[0])
        p2 = self._get_com(cv_def[1])
        p3 = self._get_com(cv_def[2])

        # get angle
        q12 = p1 - p2
        q23 = p2 - p3

        q12_n = torch.linalg.norm(q12)
        q23_n = torch.linalg.norm(q23)

        q12_u = q12 / q12_n
        q23_u = q23 / q23_n

        self.cv = torch.arccos(torch.dot(-q12_u, q23_u))

        # get forces
        if self.requires_grad:
            self.gradient = torch.autograd.grad(
                self.cv, self.coords, allow_unused=True
            )[0]
            self.gradient = self.gradient.detach().numpy()

        return float(self.cv)

    def torsion(self, cv_def: list) -> float:
        """torsion angle between four mass centers in range(-pi,pi)

        Args:
            cv_def (list):
                dihedral between atoms: [ind0, ind1, ind2, ind3]
                dihedral between center of mass: [[ind00, ind01, ...],
                                                  [ind10, ind11, ...],
                                                  [ind20, ind21, ...],
                                                  [ind30, ind 31, ...]]

        Returns:
            cv (float): computed torsional angle
        """
        if len(cv_def) != 4:
            raise ValueError(
                "CV ERROR: Invalid number of centers in definition of dihedral!"
            )

        self.update_coords()

        p1 = self._get_com(cv_def[0])
        p2 = self._get_com(cv_def[1])
        p3 = self._get_com(cv_def[2])
        p4 = self._get_com(cv_def[3])

        # get dihedral
        q12 = p2 - p1
        q23 = p3 - p2
        q34 = p4 - p3

        q23_u = q23 / torch.linalg.norm(q23)

        n1 = -q12 - torch.dot(-q12, q23_u) * q23_u
        n2 = q34 - torch.dot(q34, q23_u) * q23_u

        self.cv = torch.atan2(torch.dot(torch.cross(q23_u, n1), n2), torch.dot(n1, n2))

        # get forces
        if self.requires_grad:
            self.gradient = torch.autograd.grad(
                self.cv, self.coords, allow_unused=True
            )[0]
            self.gradient = self.gradient.detach().numpy()

        return float(self.cv)

    def linear_combination(self, cv_def: list) -> float:
        """linear combination of distances, angles or dihedrals between atoms or groups of atoms

        Args:
            cv_dev (list):
                list of distances, angle or torsions with prefactors: [[fac0, [ind00, ind01]],
                                                                       [fac1, [ind10, ind11, ind12]],
                                                                       [fac2, [ind20, ind21, ind22, ind23]],
                                                                       ...]

        Returns:
            cv (float): linear combination of distances/angles/dihedrals
        """
        self.update_coords()

        self.lc_contribs = []
        gradient = np.zeros(3 * self.natoms, dtype=float)

        for cv in cv_def:

            if len(cv[1]) == 2:
                x = self.distance(cv[1])

            elif len(cv[1]) == 3:
                x = self.angle(cv[1])

            elif len(cv[1]) == 4:
                x = self.torsion(cv[1])

            else:
                raise ValueError(
                    "CV ERROR: Invalid number of centers in definition of linear combination!"
                )

            self.lc_contribs.append(cv[0] * x)
            if self.requires_grad:
                gradient += cv[0] * self.gradient

        if self.requires_grad:
            self.gradient = gradient

        self.cv = np.asarray(self.lc_contribs).sum()
        return float(self.cv)

    def distorted_distance(self, cv_def: list, r_0: float = 3.0) -> float:
        """distorted distance between two mass centers in range(0, inf) mapped to range(1,0)

        Args:
            cv_def (list):
                distorted distance beteen atoms: [ind0, ind1]
                distorted distance between mass centers: [[ind00, ind01, ...],
                                                          [ind10, ind11, ...]]
            r_0 (float): distance in Angstrom at which the CN function has the value 0.5

        Returns:
            distorted distance (float): computed distance
        """
        self.update_coords()

        if len(cv_def) != 2:
            raise ValueError(
                "CV ERROR: Invalid number of centers in definition of distance!"
            )

        r_0 /= BOHR_to_ANGSTROM

        p1 = self._get_com(cv_def[0])
        p2 = self._get_com(cv_def[1])

        # get distance
        r12 = p2 - p1
        norm = torch.linalg.norm(r12, dtype=torch.float)
        # to prevent numerical instability
        if norm == r_0:
            d = norm / (r_0 * 1.000001)
        else:
            d = norm / r_0

        self.cv = (1.0 - torch.pow(d, 6)) / (1.0 - torch.pow(d, 12))

        # get forces
        if self.requires_grad:
            self.gradient = torch.autograd.grad(
                self.cv, self.coords, allow_unused=True
            )[0]
            self.gradient = self.gradient.detach().numpy()

        return float(self.cv)

    def rmsd(
        self,
        cv_def: Union[str, list],
        method: str = "quaternion",
    ) -> float:
        """rmsd to reference structure

        Args:
            cv_def: path to xyz file with reference structure
                    definition: 'path to xyz' or
                                ['path to reference xyz', [atom indices]]
            method: 'kabsch', 'quaternion' or 'kearsley' algorithm for optimal alignment
                    gradient of kabsch algorithm numerical unstable!

        Returns:
            cv: root-mean-square deviation to reference structure
        """
        self.update_coords()

        if isinstance(cv_def, list):
            atom_indices = cv_def[1]
            cv_def = cv_def[0]
        else:
            atom_indices = None

        if self.reference == None:
            self.reference = read_xyz(cv_def)

        if method.lower() == "kabsch":
            self.cv = kabsch_rmsd(self.coords, self.reference, indices=atom_indices)
        else:  # 'quaternion':
            self.cv = quaternion_rmsd(self.coords, self.reference, indices=atom_indices)

        if self.requires_grad:
            self.gradient = torch.autograd.grad(
                self.cv, self.coords, allow_unused=True
            )[0]
            self.gradient = self.gradient.detach().numpy()

        return float(self.cv)

    def get_cv(self, cv: str, atoms: list, **kwargs) -> Tuple[float, np.ndarray]:
        """get state of collective variable from cv definition of sampling_tools

        Args:
            cv: type of CV
            atoms: indices of atoms

        Returns:
           xi: value of collective variable
           gradient: gradient of collective variable
        """
        if cv.lower() == "x":
            xi = self.x()
            self.type = "2d"
        elif cv.lower() == "y":
            xi = self.y()
            self.type = "2d"
        elif cv.lower() == "distance":
            xi = self.distance(atoms)
            self.type = "distance"
        elif cv.lower() == "angle":
            xi = self.angle(atoms)
            self.type = "angle"
        elif cv.lower() == "torsion":
            xi = self.torsion(atoms)
            self.type = "angle"
        elif cv.lower() == "lin_comb_dists":
            xi = self.linear_combination(atoms)
            self.type = "distance"
        elif cv.lower() == "lin_comb_angles":
            xi = self.linear_combination(atoms)
            self.type = "angle"
        elif cv.lower() == "linear_combination":
            xi = self.linear_combination(atoms)
            self.type = None
        elif cv.lower() == "distorted_distance":
            xi = self.distorted_distance(atoms, **kwargs)
            self.type = None
        elif cv.lower() == "rmsd":
            xi = self.rmsd(atoms)
            self.type = "distance"
        else:
            print(" >>> Error in CV: Unknown coordinate")
            sys.exit(1)

        if self.requires_grad:
            return xi, self.gradient
        else:
            return xi



import os
import itertools
import torch
from scipy.spatial import KDTree

from adaptive_sampling.units import BOHR_to_ANGSTROM


def read_xyz(xyz_name: str) -> torch.tensor:
    """Read cartesian coordinates from file (*.xyz)

    Args:
        xyz_name (str): file-name of xyz-file

    Returns:
        mol: (3*natoms,) cartesian coordinates
    """
    if os.path.exists(xyz_name):
        xyzf = open(xyz_name, "r")
    else:
        raise ValueError(" >>> fatal error: xyz file not found")

    mol = []
    for line in xyzf:
        words = line.strip().split()
        if len(words) >= 4:
            mol.append([float(words[1]), float(words[2]), float(words[3])])

    mol = itertools.chain(*mol)
    mol = torch.FloatTensor(list(mol)) / BOHR_to_ANGSTROM
    return mol


def rmsd(V, W):
    """root-mean-square deviation"""
    diff = V - W
    return torch.sqrt(torch.sum(diff * diff) / len(V))


def kabsch_rmsd(
    coords1: torch.tensor,
    coords2: torch.tensor,
    indices: bool = None,
) -> torch.tensor:
    """minimize rmsd between cartesian coordinates by kabsch algorithm

    Args:
        coords1: (3*n_atoms,) tensor of cartesian coordinates
        coords2: (3*n_atoms,) tensor of cartesian coordinates
        indices: indices of atoms that are included

    Returns:
        coords1: (3*n_atoms,) coordinates fitted to coords2
    """
    n = len(coords1)
    coords1 = torch.reshape(coords1, (int(n / 3), 3)).float()
    coords2 = torch.reshape(coords2, (int(n / 3), 3)).float()

    if indices != None:
        coords1 = coords1[indices]
        coords2 = coords2[indices]

    # translate centroids of molecules onto each other
    coords1_new = coords1 - centroid(coords1)
    coords2_new = coords2 - centroid(coords2)

    # optimal rotation of coords1 to fit coords2
    coords1_new = kabsch_rot(coords1_new, coords2_new)
    return rmsd(coords1_new, coords2_new)


def centroid(X: torch.tensor) -> torch.tensor:
    """Centroid is the mean position of all the points in all of the coordinate
    directions, from a vectorset X.

    Args:
        X: (Natoms, 3) array of cartesian coordinates

    Returns:
        C: 3D centroid
    """
    C = X.mean(axis=0)
    return C


def kabsch_rot(coords1: torch.tensor, coords2: torch.tensor) -> torch.tensor:
    """Rotate coords1 on coords2

    Args:
        coords1: (3*n_atoms,) tensor of cartesian coordinates
        coords2: (3*n_atoms,) tensor of cartesian coordinates

    Returns:
        new_coords1: (3*n_atoms,) tensor of rotated coordinates
    """
    U = kabsch(coords1, coords2)
    coords1_new = torch.matmul(coords1, U)
    return coords1_new


def kabsch(P: torch.tensor, Q: torch.tensor):
    """Kabsch algorithm to obtain rotation matrix"""
    C = torch.matmul(torch.transpose(P, 0, 1), Q)
    V, S, W = torch.linalg.svd(C)
    d = (torch.linalg.det(V) * torch.linalg.det(W)) < 0.0

    if d:
        S[-1] = -S[-1]
        V[:, -1] = -V[:, -1]

    U = torch.matmul(V, W)
    return U


def quaternion_rmsd(
    coords1: torch.tensor,
    coords2: torch.tensor,
    indices: list = None,
    reshape: bool = True,
    return_coords: bool = False,
) -> torch.tensor:
    """
    Rotate coords1 on coords2 and calculate the RMSD
    based on doi:10.1016/1049-9660(91)90036-O

    Args:
        coords1 : (natoms*3,) tensor of cartesian coordinates
        coords3 : (natoms*3,) tensor of cartesian coordinates
        indices: list of indices that are included
        reshape: if False, coords1 not reshaped to (natoms,3)
        return_coords: if transformed coords1 are returned

    Returns:
        rmsd: minimized root-mean-square deviation
        coords1: if return_coords
    """
    if reshape:
        coords1 = torch.reshape(coords1, (int(len(coords1) / 3), 3)).float()
        if indices != None:
            coords1 = coords1[indices]

    coords2 = torch.reshape(coords2, (int(len(coords2) / 3), 3)).float()
    if indices != None:
        coords2 = coords2[indices]

    # translate centroids of molecules onto each other
    coords1_new = coords1 - centroid(coords1)
    coords2_new = coords2 - centroid(coords2)

    rot = quaternion_rotate(coords1_new, coords2_new)
    coords1_new = torch.matmul(coords1_new, rot)

    if return_coords:
        return rmsd(coords1_new, coords2_new), coords1_new
    return rmsd(coords1_new, coords2_new)


def _quaternion_transform(r: torch.tensor) -> torch.tensor:
    """Get optimal rotation"""
    Wt_r = _makeW(*r).T
    Q_r = _makeQ(*r)
    rot = torch.matmul(Wt_r, Q_r)[:3, :3]
    return rot


def _makeW(r1, r2, r3, r4=0):
    """matrix involved in quaternion rotation"""
    W = torch.tensor(
        [
            [r4, r3, -r2, r1],
            [-r3, r4, r1, r2],
            [r2, -r1, r4, r3],
            [-r1, -r2, -r3, r4],
        ]
    )
    return W


def _makeQ(r1, r2, r3, r4=0):
    """matrix involved in quaternion rotation"""
    Q = torch.tensor(
        [
            [r4, -r3, r2, r1],
            [r3, r4, -r1, r2],
            [-r2, r1, r4, r3],
            [-r1, -r2, -r3, r4],
        ]
    )
    return Q


def quaternion_rotate(X: torch.tensor, Y: torch.tensor) -> torch.tensor:
    """
    Calculate the rotation

    Args:
        X: (natoms,3) tensor of cartesian coordinates
        Y: (natoms,3) tensor of cartesian coordinates

    Returns:
        rot: Rotation matrix (3,3)
    """
    N = X.size(dim=0)
    W = torch.stack([_makeW(*Y[k]) for k in range(N)])
    Q = torch.stack([_makeQ(*X[k]) for k in range(N)])
    Qt_dot_W = torch.stack([torch.matmul(Q[k].T, W[k]) for k in range(N)])
    A = torch.sum(Qt_dot_W, axis=0)
    eigen = torch.linalg.eigh(A)
    r = eigen[1][:, eigen[0].argmax()]
    rot = _quaternion_transform(r)
    return rot



import numpy as np

# conversions
J_to_atomic = 1.0 / 4.359744e-18
atomic_to_kJmol = 2625.499639
atomic_to_K = 315775.04e0
BOHR_to_ANGSTROM = 0.52917721092e0
DEGREES_per_RADIAN = 180.0 / np.pi
atomic_to_kg = 1.66054e-27
atomic_to_fs = 1.0327503e0

# constants
kB_in_SI = 1.380648e-23
R_in_SI = 8.314
kB_in_atomic = kB_in_SI * J_to_atomic
h_in_SI = 6.62607004e-34



import numpy as np
from typing import Union, Tuple


def diff(
    a: Union[np.ndarray, float], b: Union[np.ndarray, float], cv_type: str
) -> Union[np.ndarray, float]:
    """get difference of elements of numbers or arrays
    in range(-pi, pi) if cv_type='angle' else range(-inf, inf) 

    Args:
        a: number or array
        b: number or array

    Returns:
        diff: element-wise difference (a-b)
    """
    diff = a - b

    # wrap to range(-pi,pi) for angle
    if isinstance(diff, np.ndarray) and cv_type == "angle":

        diff[diff > np.pi] -= 2 * np.pi
        diff[diff < -np.pi] += 2 * np.pi

    elif cv_type == "angle":

        if diff < -np.pi:
            diff += 2 * np.pi
        elif diff > np.pi:
            diff -= 2 * np.pi

    return diff


def sum(
    a: Union[np.ndarray, float], b: Union[np.ndarray, float], cv_type: str
) -> Union[np.ndarray, float]:
    """get sum of elements of numbers or arrays
    in range(-pi, pi) if cv_type='angle' else range(-inf, inf) 

    Args:
        a: number or array
        b: number or array

    Returns:
        diff: element-wise difference (a-b)
    """
    sum = a + b

    # wrap to range(-pi,pi) for angle
    if isinstance(diff, np.ndarray) and cv_type == "angle":

        sum[sum > np.pi] -= 2 * np.pi
        sum[sum < -np.pi] += 2 * np.pi

    elif cv_type == "angle":

        if sum < -np.pi:
            sum += 2 * np.pi
        elif sum > np.pi:
            sum -= 2 * np.pi

    return sum


def welford_var(
    count: float, mean: float, M2: float, newValue: float
) -> Tuple[float, float, float]:
    """On-the-fly estimate of sample variance by Welford's online algorithm
    
    Args:
        count: current number of samples (with new one)
        mean: current mean
        M2: helper to get variance
        newValue: new sample

    Returns:
        mean: sample mean,
        M2: sum of powers of differences from the mean
        var: sample variance
    """
    if count == 0:
        return 0.0, 0.0, 0.0
    delta = newValue - mean
    mean += delta / count
    delta2 = newValue - mean
    M2 += delta * delta2
    var = M2 / count if count > 2 else 0.0
    return mean, M2, var


def combine_welford_stats(
    count_a, 
    mean_a, 
    M2_a, 
    count_b, 
    mean_b, 
    M2_b
) -> Tuple[float, float, float, float]:
        """Combines running sample stats of welford's algorithm using Chan et al.'s algorithm.
        
        args:
            count_a, mean_a, M2_a: stats of frist subsample
            count_a, mean_a, M2_a: stats of second subsample
    
        returns:
            mean, M2 and sample variance of combined samples
        """
        count = count_a + count_b
        if count == 0:
            return 0.0, 0.0, 0.0, 0.0
        delta = mean_b - mean_a
        mean = mean_a + delta * count_b / count
        if count_b == 0:
            M2 = M2_a           
        else:
            M2 = M2_a + M2_b + (delta * delta) * ((count_a / count_b) / count)
        var = M2 / count if count > 2 else 0.0
        return count, mean, M2, var


def cond_avg(a: np.ndarray, hist: np.ndarray) -> np.ndarray:
    """get hist conditioned average of a, elements with 0 counts set to 0,

    Args:
        a: input array
        hist: histogram along cv (biased probability density)

    Returns:
        cond_avg: conditional average
    """
    return np.divide(a, hist, out=np.zeros_like(a), where=(hist != 0))
    


import math
import numpy as np
from typing import Tuple
from scipy.stats import kstat
from ..units import *


def gamd_correction_n(
    grid: np.ndarray,
    cv: np.ndarray,
    delta_V: np.ndarray,
    korder: int = 2,
    equil_temp: float = 300.0,
    return_hist: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """Correction to PMF from GaMD using kth cumulant expansion

    Args:
        grid: grid for reaction coordinate
        cv: trajectory of reaction coordinate
        delta_V: trajectory of GaMD bias potential
        korder: order of cumulant expansion, supported up to 4
        equil_temp: equilibrium temperature

    Returns:
        corr: correction to PMF from GaMD
        hist: histogram of cv on grid
    """
    if korder > 4:
        print(" >>> Warning: cumulant expansion only supportet up to 4th order")
        korder = 4

    beta = 1.0 / (R_in_SI * equil_temp / 1000.0)  # kJ/mol
    dV_kJ = delta_V * atomic_to_kJmol  # kJ/mol

    dx2 = (grid[1] - grid[0]) / 2.0
    hist = np.zeros(len(grid))
    corr = np.zeros(len(grid))
    for i, x in enumerate(grid):
        dat_i = dV_kJ[np.where(np.logical_and(cv >= x - dx2, cv < x + dx2))]
        hist[i] = len(dat_i)
        if hist[i] > 0:
            # kth order cumulant expansion
            for k in range(1, korder + 1):
                corr[i] += (np.power(beta, k) / math.factorial(k)) * kstat(dat_i, k)

    if return_hist:
        return -corr / beta, hist
    else:
        return -corr / beta


def gamd_pmf(
    grid: np.ndarray,
    cv: np.ndarray,
    delta_V: np.ndarray,
    korder: int = 2,
    equil_temp: float = 300.0,
):
    """compute pmf from GaMD simulation
    by correcting probability density with kth order cumulant expansion

    Args:
        grid: grid for reaction coordinate
        cv: trajectory of reaction coordinate
        delta_V: trajectory of GaMD bias potential
        korder: order of cumulant expansion, supported up to 4
        equil_temp: equilibrium temperature

    Returns:
        pmf: potential of mean force
        rho: probability density
    """
    RT = R_in_SI * equil_temp / 1000.0  # kJ/mol
    dx = grid[1] - grid[0]

    # compute pmf from biased prob. density corrected by second order cumulant expansion
    gamd_corr, hist = gamd_correction_n(
        grid,
        cv,
        delta_V,
        korder=korder,
        equil_temp=equil_temp,
        return_hist=True,
    )
    rho = hist / (hist.sum() * dx)
    pmf = -RT * np.log(rho, out=np.full_like(rho, np.nan), where=(rho != 0)) + gamd_corr

    rho = np.exp(-pmf / RT)
    rho /= rho[~np.isnan(pmf)].sum() * dx
    pmf = -RT * np.log(rho, out=np.full_like(rho, np.nan), where=(rho != 0))

    return pmf, rho



import numpy as np
from .enhanced_sampling import EnhancedSampling
from .utils import welford_var
from ..units import *


class GaMD(EnhancedSampling):
    """Gaussian-accelerated Molecular Dynamics

       see: Miao et. al., J. Chem. Theory Comput. (2015); https://doi.org/10.1021/acs.jctc.5b00436

       Apply an harmonic boost potential to potential energy. Independent of Collective Variable.

    Args:
        gamd_sigma0: upper limit of standard deviation of boost potential
        gamd_init_step: initial steps where no bias is applied to estimate min, max and var of potential energy
        gamd_equil_steps: equilibration steps, min, max and var of potential energy is still updated
                          force constant of coupling is calculated from previous steps
        md: Object of the MD Inteface
        cv_def: definition of the Collective Variable (CV) (see adaptive_sampling.colvars)
                [["cv_type", [atom_indices], minimum, maximum, bin_width], [possible second dimension]]
        gamd_bound: "lower": use lower bound for GaMD boost
                    "upper: use upper bound for GaMD boost
        confine: is system should be confined to range of CV
        equil_temp: equillibrium temperature of MD
        verbose: print verbose information
        kinetice: calculate necessary data to obtain kinetics of reaction
        f_conf: force constant for confinement of system to the range of interest in CV space
        output_freq: frequency in steps for writing outputs

    """

    def __init__(
        self,
        gamd_sigma0: float,
        gamd_init_steps: int,
        gamd_equil_steps: int,
        *args,
        gamd_bound: str = "lower",
        confine: bool = True,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.sigma0 = gamd_sigma0
        self.gamd_init_steps = gamd_init_steps
        self.gamd_equil_steps = gamd_equil_steps
        self.gamd_bound = gamd_bound.lower()
        self.confine = confine

        self.pot_count = 0
        self.pot_var = 0.0
        self.pot_std = 0.0
        self.pot_m2 = 0.0
        self.pot_avg = 0.0
        self.pot_min = 0.0
        self.pot_max = 0.0
        self.k0 = 0.0
        self.E = 0.0

        self.gamd_pot = 0.0
        self.gamd_pot_traj = []

        self.gamd_c1 = np.zeros_like(self.histogram)
        self.gamd_c2 = np.zeros_like(self.histogram)
        self.gamd_m2 = np.zeros_like(self.histogram)
        self.gamd_corr = np.zeros_like(self.histogram)

    def step_bias(self, write_output: bool = True, write_traj: bool = True, **kwargs):

        md_state = self.the_md.get_sampling_data()
        (xi, delta_xi) = self.get_cv(**kwargs)

        # get energy and forces
        bias_force = np.zeros_like(md_state.forces)
        self.gamd_forces = np.copy(md_state.forces)
        epot = md_state.epot

        if md_state.step < self.gamd_init_steps:
            self._update_pot_distribution(epot)

        else:
            if md_state.step == self.gamd_init_steps:
                self._calc_E_k0()

            # apply boost potential
            prefac = self.k0 / (self.pot_max - self.pot_min)
            self.gamd_pot = 0.5 * prefac * np.power(self.E - epot, 2)
            bias_force -= prefac * (self.E - epot) * self.gamd_forces

            if md_state.step < self.gamd_equil_steps:
                self._update_pot_distribution(epot)
                self._calc_E_k0()

            else:
                # free energy reweighting in production
                if self._check_boundaries(xi):

                    bink = self.get_index(xi)
                    self.histogram[bink[1], bink[0]] += 1

                    # first and second order cumulants for free energy reweighting
                    (
                        self.gamd_c1[bink[1], bink[0]],
                        self.gamd_m2[bink[1], bink[0]],
                        self.gamd_c2[bink[1], bink[0]],
                    ) = welford_var(
                        self.histogram[bink[1], bink[0]],
                        self.gamd_c1[bink[1], bink[0]],
                        self.gamd_m2[bink[1], bink[0]],
                        self.gamd_pot,
                    )

        if self.confine:
            bias_force += self.harmonic_walls(xi, delta_xi)

        self.traj = np.append(self.traj, [xi], axis=0)
        self.temp.append(md_state.temp)
        self.epot.append(md_state.epot)
        self.gamd_pot_traj.append(self.gamd_pot)

        # correction for kinetics
        if self.kinetics:
            self._kinetics(delta_xi)

        if md_state.step % self.out_freq == 0:
            # write output

            if write_traj:
                self.write_traj()

            if write_output:
                self.get_pmf()
                output = {
                    "hist": self.histogram,
                    "free energy": self.pmf,
                    "gamd_corr": self.gamd_corr,
                }
                self.write_output(output, filename="gamd.out")
                self.write_restart()

        return bias_force

    def get_pmf(self):

        kBT = kB_in_atomic * self.equil_temp
        self.gamd_corr = -self.gamd_c1 - self.gamd_c2 / (2.0 * kBT)

        self.pmf = -kBT * np.log(
            self.histogram,
            out=np.zeros_like(self.histogram),
            where=(self.histogram != 0),
        )
        self.pmf += self.gamd_corr
        self.pmf *= atomic_to_kJmol
        self.pmf -= self.pmf.min()

    def shared_bias(self):
        """TODO: fix me"""
        pass

    def _update_pot_distribution(self, epot: float):
        """update min, max, avg, var and std of epot

        Args:
            epot: potential energy
        """
        self.pot_min = np.min([epot, self.pot_min])
        self.pot_max = np.max([epot, self.pot_max])
        self.pot_count += 1
        self.pot_avg, self.pot_m2, self.pot_var = welford_var(
            self.pot_count, self.pot_avg, self.pot_m2, epot
        )
        self.pot_std = np.sqrt(self.pot_var)

    def _calc_E_k0(self):
        """compute force constant for gamd boost potential

        Args:
            epot: potential energy
        """
        if self.gamd_bound == "lower":
            self.E = self.pot_max
            ko = (self.sigma0 / self.pot_std) * (
                (self.pot_max - self.pot_min) / (self.pot_max - self.pot_avg)
            )
            self.k0 = np.min([1.0, ko])

        elif self.gamd_bound == "upper":
            ko = (1.0 - self.sigma0 / self.pot_std) * (
                (self.pot_max - self.pot_min) / (self.pot_max - self.pot_avg)
            )
            if 0.0 < ko <= 1.0:
                self.k0 = ko
            else:
                self.k0 = 1.0
            self.E = self.pot_min + (self.pot_max - self.pot_min) / self.k0

        else:
            raise ValueError(f" >>> Error: unknown GaMD bound {self.gamd_bound}!")

    def write_restart(self, filename: str = "restart_gamd"):
        """write restart file
        TODO: fix me

        Args:
            filename: name of restart file
        """
        self._write_restart(
            filename=filename,
            hist=self.histogram,
            pmf=self.pmf,
            c1=self.gamd_c1,
            m2=self.gamd_m2,
            corr=self.gamd_corr,
            pot_count=self.pot_count,
            pot_var=self.pot_var,
            pot_std=self.pot_std,
            pot_m2=self.pot_m2,
            pot_avg=self.pot_avg,
            pot_min=self.pot_min,
            pot_max=self.pot_max,
            k0=self.k0,
        )

    def restart(self, filename: str = "restart_gamd"):
        """restart from restart file

        Args:
            filename: name of restart file
        """
        try:
            data = np.load(filename + ".npz", allow_pickle=True)
        except:
            raise OSError(f" >>> fatal error: restart file {filename}.npz not found!")

        self.histogram = data["hist"]
        self.pmf = data["pmf"]
        self.gamd_c1 = data["c1"]
        self.gamd_m2 = data["m2"]
        self.gamd_corr = data["corr"]
        self.pot_count = data["pot_count"]
        self.pot_var = data["pot_var"]
        self.pot_std = data["pot_std"]
        self.pot_m2 = data["pot_m2"]
        self.pot_avg = data["pot_avg"]
        self.pot_min = data["pot_min"]
        self.pot_max = data["pot_max"]
        self.k0 = data["k0"]

        if self.verbose:
            print(f" >>> Info: Adaptive sampling restartet from {filename}!")

    def write_traj(self):
        """save trajectory for post-processing"""
        data = {
            "E gamd [H]": self.gamd_pot_traj,
            "E pot [H]": self.epot,
            "T [K]": self.temp,
        }
        self._write_traj(data)

        # reset trajectories to save memory
        self.traj = np.array([self.traj[-1]])
        self.gamd_pot_traj = [self.gamd_pot_traj[-1]]
        self.epot = [self.epot[-1]]
        self.temp = [self.temp[-1]]



import random, os, itertools, time
import numpy as np
from typing import Union
from .enhanced_sampling import EnhancedSampling
from .abf import ABF
from .utils import welford_var, combine_welford_stats, diff, cond_avg
from ..processing_tools.thermodynamic_integration import integrate
from ..units import *


class eABF(ABF, EnhancedSampling):
    """Extended-system adaptive biasing Force method

       see: Lesage et. al., J. Phys. Chem. B (2017); https://doi.org/10.1021/acs.jpcb.6b10055

    The collective variable is coupled to a fictitious particle with an harmonic force.
    The dynamics of the fictitious particel is biased with the ABF algorithm.

    Args:
        ext_sigma: thermal width of coupling between collective and extended variable
        ext_mass: mass of extended variable in atomic units
        md: Object of the MDInterface
        cv_def: definition of the Collective Variable (CV) (see adaptive_sampling.colvars)
                [["cv_type", [atom_indices], minimum, maximum, bin_width], [possible second dimension]]
        nfull: Number of force samples per bin where full bias is applied,
               if nsamples < nfull the bias force is scaled down by nsamples/nfull
        friction: friction coefficient for Lagevin dynamics of the extended-system
        seed_in: random seed for Langevin dynamics of extended-system
        equil_temp: equillibrium temperature of MD
        verbose: print verbose information
        kinetice: calculate necessary data to obtain kinetics of reaction
        f_conf: force constant for confinement of system to the range of interest in CV space
        output_freq: frequency in steps for writing outputs
    """

    def __init__(
        self,
        ext_sigma: Union[float, list],
        ext_mass: Union[float, list],
        *args,
        friction: float = 1.0e-3,
        seed_in: int = 42,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        ext_sigma = [ext_sigma] if not hasattr(ext_sigma, "__len__") else ext_sigma
        ext_mass = [ext_mass] if not hasattr(ext_mass, "__len__") else ext_mass

        (xi, _) = self.get_cv()

        # for dynamics of extended-system
        self.ext_sigma = self.unit_conversion_cv(np.asarray(ext_sigma))[0]
        self.ext_k = (kB_in_atomic * self.equil_temp) / (
            self.ext_sigma * self.ext_sigma
        )
        self.ext_mass = np.asarray(ext_mass)
        self.ext_hist = np.zeros_like(self.histogram)
        self.ext_forces = np.zeros(self.ncoords)
        self.ext_momenta = np.zeros(self.ncoords)
        self.ext_coords = np.copy(xi)
        self.correction_czar = np.zeros_like(self.bias)
        self.czar_force = np.zeros_like(self.bias)
        self.friction = friction
        self.ext_traj = np.copy(self.traj)

        # set random seed for langevin dynamics
        if type(seed_in) is int:
            random.seed(seed_in)
        else:
            try:
                random.setstate(seed_in)
            except:
                if self.verbose:
                    print(
                        "\n >>> Warning: The provided seed was neither an int nor a state of random!\n"
                    )

        # initialize extended system at target temp of MD simulation
        for i in range(self.ncoords):
            self.ext_momenta[i] = random.gauss(0.0, 1.0) * np.sqrt(
                self.equil_temp * self.ext_mass[i]
            )
            ttt = (np.power(self.ext_momenta, 2) / self.ext_mass).sum()
            ttt /= self.ncoords
            self.ext_momenta *= np.sqrt(self.equil_temp / (ttt * atomic_to_K))

    def step_bias(self, write_output: bool = True, write_traj: bool = True, **kwargs):

        md_state = self.the_md.get_sampling_data()
        (xi, delta_xi) = self.get_cv(**kwargs)
        self._propagate()

        bias_force = self._extended_dynamics(xi, delta_xi)  # , self.ext_sigma)
        force_sample = [0 for _ in range(2 * self.ncoords)]
        
        if self._check_boundaries(self.ext_coords):

            bin_la = self.get_index(self.ext_coords)
            self.ext_hist[bin_la[1], bin_la[0]] += 1

            for i in range(self.ncoords):

                # linear ramp function
                ramp = (
                    1.0
                    if self.ext_hist[bin_la[1], bin_la[0]] > self.nfull
                    else self.ext_hist[bin_la[1], bin_la[0]] / self.nfull
                )

                # apply bias force on extended system
                force_sample[i] = self.ext_k[i] * diff(self.ext_coords[i], xi[i], self.cv_type[i])
                (
                    self.bias[i][bin_la[1], bin_la[0]],
                    self.m2_force[i][bin_la[1], bin_la[0]],
                    self.var_force[i][bin_la[1], bin_la[0]],
                ) = welford_var(
                    self.ext_hist[bin_la[1], bin_la[0]],
                    self.bias[i][bin_la[1], bin_la[0]],
                    self.m2_force[i][bin_la[1], bin_la[0]],
                    force_sample[i],
                )
                self.ext_forces -= ramp * self.bias[i][bin_la[1], bin_la[0]]

        # xi-conditioned accumulators for CZAR
        if self._check_boundaries(xi):

            bink = self.get_index(xi)
            self.histogram[bink[1], bink[0]] += 1

            for i in range(self.ncoords):
                force_sample[self.ncoords+i] = self.ext_k[i] * diff(
                    self.ext_coords[i], self.grid[i][bink[i]], self.cv_type[i]
                )
                self.correction_czar[i][bink[1], bink[0]] += force_sample[self.ncoords+i]

        # shared-bias eABF
        if self.shared:                 
            self.shared_bias(
                xi,
                force_sample, 
                **kwargs,
            )

        self.traj = np.append(self.traj, [xi], axis=0)
        self.ext_traj = np.append(self.ext_traj, [self.ext_coords], axis=0)
        self.temp.append(md_state.temp)
        self.epot.append(md_state.epot)
        self._up_momenta()

        # correction for kinetics
        if self.kinetics:
            self._kinetics(delta_xi)

        if md_state.step % self.out_freq == 0:
            # write output

            if write_traj:
                self.write_traj()

            if write_output:
                self.get_pmf()
                output = {"hist": self.histogram, "free energy": self.pmf}
                for i in range(self.ncoords):
                    output[f"bias force {i}"] = self.bias[i]
                    output[f"var force {i}"] = self.var_force[i]
                    output[f"czar force {i}"] = self.czar_force[i]

                self.write_output(output, filename="eabf.out")
                self.write_restart()

        return bias_force

    def get_pmf(self, method: str = "trapezoid"):

        log_rho = np.log(
            self.histogram,
            out=np.zeros_like(self.histogram),
            where=(0 != self.histogram),
        )
        avg_force = cond_avg(self.correction_czar, self.histogram)

        if self.ncoords == 1:
            self.czar_force[0] = (
                -kB_in_atomic * self.equil_temp * np.gradient(log_rho[0], self.grid[0])
                + avg_force[0]
            )
            self.pmf[0, :], _ = integrate(
                self.czar_force[0][0],
                self.dx,
                equil_temp=self.equil_temp,
                method=method,
            )
            self.pmf *= atomic_to_kJmol
            self.pmf -= self.pmf.min()

        else:
            der_log_rho = np.gradient(log_rho, self.grid[1], self.grid[0])
            self.czar_force[0] = (
                -kB_in_atomic * self.equil_temp * der_log_rho[1] + avg_force[0]
            )
            self.czar_force[1] = (
                -kB_in_atomic * self.equil_temp * der_log_rho[0] + avg_force[1]
            )
            if self.verbose:
                print(
                    " >>> Info: On-the-fly integration only available for 1D coordinates"
                )

    def shared_bias(
        self, 
        xi,
        force_sample,
        sync_interval: int=50,
        mw_file: str="../shared_bias",
        n_trials: int=10,
    ):
        """syncs eABF bias with other walkers

        TODO: 2D collective variables

        Args:
            xi: state of the collective variable
            force_sample: force sample of current step
            sync_interval: number of steps between sychronisation
            mw_file: name of buffer file for shared-bias
            n_trials: number of attempts to access of buffer file before throwing an error
        """
        md_state = self.the_md.get_sampling_data()
        if md_state.step == 0:        
            if self.verbose:
                print(" >>> Info: Creating a new instance for shared-bias eABF.")
                print(" >>> Info: Data of local walker stored in `restart_eabf_local.npz`.")
            
            # create seperate restart file with local data only
            self._write_restart(
                filename="restart_eabf_local",
                hist=self.histogram,
                force=self.bias,
                var=self.var_force,
                m2=self.m2_force,
                ext_hist=self.ext_hist,
                czar_corr=self.correction_czar,
            )
            self.update_samples = np.zeros(shape=(sync_interval, len(force_sample)))
            self.last_samples_xi = np.zeros(shape=(sync_interval, len(xi)))
            self.last_samples_la = np.zeros(shape=(sync_interval, len(self.ext_coords)))

            if not os.path.isfile(mw_file+".npz"):
                if self.verbose:
                    print(f" >>> Info: Creating buffer file for shared-bias eABF: `{mw_file}.npz`.")
                self._write_restart(
                    filename=mw_file,
                    hist=self.histogram,
                    force=self.bias,
                    var=self.var_force,
                    m2=self.m2_force,
                    ext_hist=self.ext_hist,
                    czar_corr=self.correction_czar,
                )
                os.chmod(mw_file + ".npz", 0o444)
            elif self.verbose:
                print(f" >>> Info: Syncing with existing buffer file for shared-bias eABF: `{mw_file}.npz`.")
        
        count = md_state.step % sync_interval
        self.update_samples[count] = force_sample
        self.last_samples_xi[count] = xi
        self.last_samples_la[count] = self.ext_coords

        if count == sync_interval-1:
            
            # calculate progress since last sync from new samples
            hist = np.zeros_like(self.histogram)
            m2 = np.zeros_like(self.m2_force)
            bias = np.zeros_like(self.bias)
            ext_hist = np.zeros_like(self.ext_hist)
            czar_corr = np.zeros_like(self.correction_czar)
            
            for i, sample in enumerate(self.update_samples):
                if self._check_boundaries(self.last_samples_la[i]):                   
                    bin_la = self.get_index(self.last_samples_la[i])
                    ext_hist[bin_la[1], bin_la[0]] += 1
                    for j in range(self.ncoords):
                        (
                            bias[j][bin_la[1], bin_la[0]],
                            m2[j][bin_la[1], bin_la[0]],
                            _,
                        ) = welford_var(
                            ext_hist[bin_la[1], bin_la[0]],
                            bias[j][bin_la[1], bin_la[0]],
                            m2[j][bin_la[1], bin_la[0]],
                            sample[j],
                        )
                
                if self._check_boundaries(self.last_samples_xi[i]):
                    bin_xi = self.get_index(self.last_samples_xi[i])
                    hist[bin_xi[1], bin_xi[0]] += 1 
                    for j in range(self.ncoords):
                        czar_corr[j][bin_xi[1], bin_xi[0]] += sample[self.ncoords+j]
            
            
            self._update_eabf(
                "restart_eabf_local",
                hist, 
                ext_hist,
                bias, 
                m2,
                czar_corr,
            )

            trial = 0
            while trial < n_trials:
                trial += 1
                if not os.access(mw_file + ".npz", os.W_OK):
                    
                    # grant write access only to one walker during sync
                    os.chmod(mw_file + ".npz", 0o666) 
                    self._update_eabf(
                        mw_file,
                        hist, 
                        ext_hist,
                        bias, 
                        m2,
                        czar_corr,
                    )
                    self.restart(filename=mw_file)
                    os.chmod(mw_file + ".npz", 0o444)  # other walkers can access again
                    
                    self.get_pmf()  # get new global pmf   
                    break                     

                elif trial < n_trials:
                    if self.verbose:
                        print(f" >>> Warning: Retry to open shared buffer file after {trial} failed attempts.")
                    time.sleep(0.1)
                else:
                    raise Exception(f" >>> Fatal Error: Failed to sync bias with `{mw_file}.npz`.")

    def _propagate(self, langevin: bool = True):
        """Propagate momenta/coords of extended variable in time with Velocity Verlet

        Args:
           langevin: Temperature control with langevin dynamics
        """
        if langevin:
            prefac = 2.0 / (2.0 + self.friction * self.the_md.dt)
            rand_push = np.sqrt(
                self.equil_temp * self.friction * self.the_md.dt * kB_in_atomic / 2.0e0
            )
            self.ext_rand_gauss = np.zeros(shape=(len(self.ext_momenta),), dtype=float)
            for atom in range(len(self.ext_rand_gauss)):
                self.ext_rand_gauss[atom] = random.gauss(0, 1)

            self.ext_momenta += np.sqrt(self.ext_mass) * rand_push * self.ext_rand_gauss
            self.ext_momenta -= 0.5e0 * self.the_md.dt * self.ext_forces
            self.ext_coords += (
                prefac * self.the_md.dt * self.ext_momenta / self.ext_mass
            )

        else:
            self.ext_momenta -= 0.5e0 * self.the_md.dt * self.ext_forces
            self.ext_coords += self.the_md.dt * self.ext_momenta / self.ext_mass

    def _up_momenta(self, langevin: bool = True):
        """Update momenta of extended variables with Velocity Verlet

        Args:
            langevin: Temperature control with langevin dynamics
        """
        if langevin:
            prefac = (2.0e0 - self.friction * self.the_md.dt) / (
                2.0e0 + self.friction * self.the_md.dt
            )
            rand_push = np.sqrt(
                self.equil_temp * self.friction * self.the_md.dt * kB_in_atomic / 2.0e0
            )
            self.ext_momenta *= prefac
            self.ext_momenta += np.sqrt(self.ext_mass) * rand_push * self.ext_rand_gauss
            self.ext_momenta -= 0.5e0 * self.the_md.dt * self.ext_forces
        else:
            self.ext_momenta -= 0.5e0 * self.the_md.dt * self.ext_forces

    def _extended_dynamics(
        self,
        xi: np.ndarray,
        delta_xi: np.ndarray,
        margin: np.ndarray = np.array([0.0, 0.0]),
    ) -> np.ndarray:
        """get forces due to coupling to extended system and confine extended variable to range of interest

        Args:
            xi: collective variable
            delta_xi: gradient of collective variable
            margin: inset from minx, maxx where bias starts

        Returns:
            bias_force:
        """
        bias_force = np.zeros_like(self.the_md.forces)

        for i in range(self.ncoords):
            # harmonic coupling of extended coordinate to reaction coordinate

            dxi = diff(self.ext_coords[i], xi[i], self.cv_type[i])
            self.ext_forces[i] = self.ext_k[i] * dxi
            bias_force -= self.ext_k[i] * dxi * delta_xi[i]

            # harmonic walls for confinement to range of interest
            if self.ext_coords[i] > (self.maxx[i] - margin[i]):
                r = diff(self.maxx[i] - margin[i], self.ext_coords[i], self.cv_type[i])
                self.ext_forces[i] -= self.f_conf[i] * r

            elif self.ext_coords[i] < (self.minx[i] + margin[i]):
                r = diff(self.minx[i] + margin[i], self.ext_coords[i], self.cv_type[i])
                self.ext_forces[i] -= self.f_conf[i] * r

        return bias_force

    def _update_eabf(
        self, 
        filename: str,
        hist, 
        ext_hist,
        bias, 
        m2,
        czar_corr,
    ):
        with np.load(f"{filename}.npz") as data:  

            new_hist = data["ext_hist"] + hist
            new_czar_corr = data["czar_corr"] + czar_corr
            new_ext_hist = np.zeros_like(self.histogram).flatten()
            new_m2 = np.zeros_like(self.m2_force).flatten()
            new_bias = np.zeros_like(self.bias).flatten() 

            for i in range(len(new_ext_hist)):
                (
                    new_ext_hist[i],
                    new_bias[i],
                    new_m2[i],
                    _,
                ) = combine_welford_stats(
                    data["ext_hist"].flatten()[i], 
                    data["force"].flatten()[i], 
                    data["m2"].flatten()[i], 
                    ext_hist.flatten()[i], 
                    bias.flatten()[i], 
                    m2.flatten()[i],
                )

        self._write_restart(
            filename=filename,
            hist=new_hist,
            force=new_bias.reshape(self.bias.shape),
            m2=new_m2.reshape(self.m2_force.shape),
            ext_hist=new_ext_hist.reshape(self.ext_hist.shape),
            czar_corr=new_czar_corr,
        ) 

    def write_restart(self, filename: str = "restart_abf"):
        """write restart file

        Args:
            filename: name of restart file
        """
        self._write_restart(
            filename=filename,
            hist=self.histogram,
            force=self.bias,
            m2=self.m2_force,
            ext_hist=self.ext_hist,
            czar_corr=self.correction_czar,
        )

    def restart(self, filename: str = "restart_abf"):
        """restart from restart file

        Args:
            filename: name of restart file
        """
        try:
            data = np.load(filename + ".npz")
        except:
            raise OSError(f" >>> Fatal Error: restart file `{filename}.npz` not found!")

        self.histogram = data["hist"]
        self.bias = data["force"]
        self.m2_force = data["m2"]
        self.ext_hist = data["ext_hist"]
        self.correction_czar = data["czar_corr"]

        if self.verbose:
            print(f" >>> Info: Adaptive sampling restartet from {filename}!")

    def _write_ext_traj(self):
        data = {}
        for i in range(self.ncoords):
            if self.cv_type[i] == "angle":
                self.ext_traj[:, i] *= DEGREES_per_RADIAN
            elif self.cv_type[i] == "distance":
                self.ext_traj[:, i] *= BOHR_to_ANGSTROM
            data[f"lambda{i}"] = self.ext_traj[:, i]
        return data

    def write_traj(self):
        """save trajectory for post-processing"""

        data = self._write_ext_traj()
        data["Epot [H]"] = self.epot
        data["T [K]"] = self.temp

        self._write_traj(data)

        # reset trajectories to save memory
        self.traj = np.array([self.traj[-1]])
        self.ext_traj = np.array([self.ext_traj[-1]])
        self.epot = [self.epot[-1]]
        self.temp = [self.temp[-1]]



import numpy as np
from typing import Union, List, Tuple
from ..units import *


def rolling_avg(traj: np.ndarray, n: int = 50) -> np.ndarray:
    """calculates the rolling average of a time trajectory"""
    return np.convolve(traj, np.ones(n), "valid") / n


def harmonic(
    x: Union[float, np.ndarray], x0: Union[float, np.ndarray], kx: float
) -> Union[float, np.ndarray]:
    """harmonic potential as implemented in the adaptive biasing module"""
    return 0.5 * kx * (np.power((x - x0), 2))


def harmonic_force(
    x: Union[float, np.ndarray], x0: Union[float, np.ndarray], kx: float
) -> Union[float, np.ndarray]:
    """harmonic force"""
    return kx * (x - x0)


def boltzmann(u_pot: Union[float, np.ndarray], beta: float) -> Union[float, np.ndarray]:
    """get Boltzmann factors"""
    return np.exp(-beta * u_pot, dtype=float)


def join_frames(traj_list: List[np.array]) -> Tuple[np.ndarray, float, np.ndarray]:
    """get one array with all frames from list of trajectories

    Args:
        traj_list: list of trajectories

    Returns:
        all_frames: array with all samples
        num_frames: total numper of samples
        frames_per_traj: array with number of samples per original trajectorie
    """
    num_frames = 0
    frames_per_traj = []
    all_frames = traj_list[0]
    for ii, traj in enumerate(traj_list):
        frames_per_traj.append(len(traj))
        num_frames += frames_per_traj[ii]
        if ii > 0:
            all_frames = np.concatenate((all_frames, traj))

    frames_per_traj = np.array(frames_per_traj)
    return all_frames, num_frames, frames_per_traj


def welford_var(
    count: float, mean: float, M2: float, newValue: float
) -> Tuple[float, float, float]:
    """On-the-fly estimate of sample variance by Welford's online algorithm

    Args:
        count: current number of samples (with new one)
        mean: current mean
        M2: helper to get variance
        newValue: new sample

    Returns:
        mean (float)
        M2 (float)
        var (float)
    """
    delta = newValue - mean
    mean += delta / count
    delta2 = newValue - mean
    M2 += delta * delta2
    var = M2 / count if count > 2 else 0.0
    return mean, M2, var


def weighted_hist(grid: np.ndarray, obs: np.ndarray, weights: np.ndarray) -> np.ndarray:
    """get weighted histogram of data

    Args:
        grid: bins along reaction coordinate
        obs: data frames of observable
        weights: weights of data frames

    Returns:
        hist: weighted histogram
    """
    weights /= weights.sum()
    hist = np.zeros(len(grid))
    dx = grid[1] - grid[0]
    idx = np.arange(0, len(obs), 1)
    for i, x in enumerate(grid):
        dat_x = idx[np.where(np.logical_and(obs >= x - dx / 2.0, obs <= x + dx / 2.0))]
        hist[i] = (obs[dat_x] * weights[dat_x]).sum()
    hist /= hist.sum() * dx
    return hist


def ensemble_average(obs: np.ndarray, weights: np.ndarray) -> tuple:
    """ensemble average of observable

    Args:
        obs (np.ndarray): trajectory of observables
        weights (np.ndarray): weigths of data frames

    Returns:
        avg (float): ensemble average
        sem (float): standard error of the mean of avg
    """
    weights /= weights.sum()
    avg = np.average(obs, weights=weights)  # ensemble average
    std = np.sqrt(
        np.average(np.power(obs - avg, 2), weights=weights)
    )  # standard deviation
    sem = std / np.sqrt(float(len(obs)))  # standard error of the mean
    return avg, sem


def conditional_average(
    grid: np.ndarray, xi_traj: np.ndarray, obs: np.ndarray, weights: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """bin wise conditional average and standard error of the mean

    Args:
        grid: bins along reaction coordinate
        xi_traj: trajectory of reaction coordinate
        obs: trajectory of observables
        weights: weigths of data frames

    Returns:
        obs_xi: conditional average along reaction coordinate
        sem_xi: standard error of the mean of obs_xi
    """
    weights /= weights.sum()
    dx = grid[1] - grid[0]
    dx2 = dx / 2

    obs_xi = np.zeros(len(grid))
    sem_xi = np.zeros(len(grid))

    for i, x in enumerate(grid):
        dat_i = np.where(np.logical_and(xi_traj >= x - dx2, xi_traj < x + dx2))
        nsamples = len(obs[dat_i])
        if nsamples > 0:
            p_i = weights[dat_i] / weights[dat_i].sum()
            obs_xi[i] = np.sum(p_i * obs[dat_i])
            sem_xi[i] = np.sqrt(
                np.sum(p_i * np.square(obs[dat_i] - obs_xi[i]))
            ) / np.sqrt(nsamples)
    return obs_xi, sem_xi


def reaction_freeE(
    pmf: np.ndarray,
    T: float = 300.0,
    min_bin: int = 20,
    max_bin: int = -20,
    TS: int = None,
) -> tuple:
    """calculate free energy difference
       see: Dietschreit et al., J. Chem. Phys. 156, 114105 (2022); https://doi.org/10.1063/5.0083423

    Args:
        pmf: potential of mean force (free energy surface)
        T: temperature
        min_bin/max_bin: minimum/maximum bin for search of transition state
        TS: alternatively, bin number of TS

    Returns:
        dA (float): free energy difference
        dA_grid (np.ndarray): free energy difference on grid
    """
    RT = (R_in_SI * T) / 1000.0
    pmf = pmf[~np.isnan(pmf)]

    if TS == None:
        TS = np.where(pmf == np.amax(pmf[min_bin:max_bin]))[0][0]

    P = np.exp(-pmf / RT)
    P /= P.sum()

    P_a = P[:TS].sum()
    P_b = P[(TS + 1) :].sum()

    dA = RT * np.log(P_a / P_b)
    dA_grid = np.zeros(len(pmf))
    dA_grid[TS] += dA / 2
    dA_grid[(TS + 1) :] += dA
    dA_approx = pmf[TS:].min() - pmf[:TS].min()

    return dA, dA_grid, dA_approx


def activation_freeE(
    pmf: np.ndarray,
    m_xi_inv: np.ndarray,
    T: float = 300.0,
    min_bin: int = 20,
    max_bin: int = -20,
    TS: int = None,
) -> tuple:
    """calculate activation free energy
       see: Dietschreit et al., J. Chem. Phys., 157, 084113 (2022).; <https://aip.scitation.org/doi/10.1063/5.0102075>

    Args:
        pmf: potential of mean force (free energy surface)
        m_xi_inv: z-conditioned average of inverse mass associates with CV, expected units are xi^2/(au_mass * angstrom^2)
        T: temperature
        min_bin/max_bin: minimum/maximum bin for search of transition state
        TS: alternatively, bin number of TS

    Returns:
        dA (float): free energy difference
        dA_grid (np.ndarray): free energy difference on grid
    """
    RT = (R_in_SI * T) / 1000.0  # kJ/mol

    pmf = pmf[~np.isnan(pmf)]
    if TS == None:
        TS = np.where(pmf == np.amax(pmf[min_bin:max_bin]))[0][0]

    rho = np.exp(-pmf / RT)
    P = rho / rho.sum()  # normalize so that P_a + P_b = 1.0

    lambda_xi = np.sqrt(
        (h_in_SI * h_in_SI * m_xi_inv[TS]) / (2.0 * np.pi * atomic_to_kg * kB_in_SI * T)
    )
    lambda_xi *= 1e10

    P_a = P[:TS].sum()
    P_b = P[(TS + 1) :].sum()

    dA_a2b = -RT * np.log((rho[TS] * lambda_xi) / P_a)
    dA_b2a = -RT * np.log((rho[TS] * lambda_xi) / P_b)
    dA_a2b_approx = pmf[TS] - pmf[:TS].min()
    dA_b2a_approx = pmf[TS] - pmf[TS:].min()

    return dA_a2b, dA_b2a, dA_a2b_approx, dA_b2a_approx



import os, time, itertools
import numpy as np
from .enhanced_sampling import EnhancedSampling
from .utils import welford_var, combine_welford_stats
from ..processing_tools.thermodynamic_integration import integrate
from ..units import *


class ABF(EnhancedSampling):
    """Adaptive Biasing Force Method

       see: Comer et. al., J. Phys. Chem. B (2015); https://doi.org/10.1021/jp506633n

    Args:
        md: Object of the MD Interface
        cv_def: definition of the Collective Variable (CV) (see adaptive_sampling.colvars)
                [["cv_type", [atom_indices], minimum, maximum, bin_width], [possible second dimension]]
        nfull: Number of force samples per bin where full bias is applied,
               if nsamples < nfull the bias force is scaled down by nsamples/nfull
        equil_temp: equillibrium temperature of MD
        verbose: print verbose information
        kinetice: calculate necessary data to obtain kinetics of reaction
        f_conf: force constant for confinement of system to the range of interest in CV space
        output_freq: frequency in steps for writing outputs
    """

    def __init__(self, *args, nfull: int = 100, **kwargs):
        super().__init__(*args, **kwargs)
        self.nfull = nfull
        self.var_force = np.zeros_like(self.bias)
        self.m2_force = np.zeros_like(self.bias)

    def step_bias(self, write_output: bool = True, write_traj: bool = True, **kwargs):

        md_state = self.the_md.get_sampling_data()

        (xi, delta_xi) = self.get_cv(**kwargs)

        bias_force = np.zeros_like(md_state.forces)

        if self._check_boundaries(xi):

            bink = self.get_index(xi)
            self.histogram[bink[1], bink[0]] += 1
            force_sample = [0 for _ in range(self.ncoords)]
            for i in range(self.ncoords):

                # linear ramp function
                ramp = (
                    1.0
                    if self.histogram[bink[1], bink[0]] > self.nfull
                    else self.histogram[bink[1], bink[0]] / self.nfull
                )

                # inverse gradient v_i
                delta_xi_n = np.linalg.norm(delta_xi[i])
                v_i = delta_xi[i] / (delta_xi_n * delta_xi_n)

                # apply bias force
                force_sample[i] = np.dot(
                    md_state.forces, v_i
                ) - kB_in_atomic * self.equil_temp * self._divergence_xi(
                    xi[i], self.cv_type[i]
                )

                (
                    self.bias[i][bink[1], bink[0]],
                    self.m2_force[i][bink[1], bink[0]],
                    self.var_force[i][bink[1], bink[0]],
                ) = welford_var(
                    self.histogram[bink[1], bink[0]],
                    self.bias[i][bink[1], bink[0]],
                    self.m2_force[i][bink[1], bink[0]],
                    force_sample[i],
                )

                bias_force -= ramp * self.bias[i][bink[1], bink[0]] * delta_xi[i]
            
            if self.shared:
                self.shared_bias(list(itertools.chain(*[bink, force_sample])), **kwargs)
        
        else:
            bias_force += self.harmonic_walls(xi, delta_xi)

        self.traj = np.append(self.traj, [xi], axis=0)
        self.temp.append(md_state.temp)
        self.epot.append(md_state.epot)

        # correction for kinetics
        if self.kinetics:
            self._kinetics(delta_xi)

        if md_state.step % self.out_freq == 0:
            # write output

            if write_traj:
                self.write_traj()

            if write_output:
                self.get_pmf()
                output = {"hist": self.histogram, "free energy": self.pmf}
                for i in range(self.ncoords):
                    output[f"bias force {i}"] = self.bias[i]
                    output[f"var force {i}"] = self.var_force[i]

                self.write_output(output, filename="abf.out")
                self.write_restart()

        return bias_force

    def get_pmf(self, method: str = "trapezoid"):

        if self.ncoords == 1:
            self.pmf[0, :], _ = integrate(
                self.bias[0][0], self.dx, equil_temp=self.equil_temp, method=method
            )
            self.pmf *= atomic_to_kJmol
            self.pmf -= self.pmf.min()

        elif self.verbose:
            print(" >>> Info: On-the-fly integration only available for 1D coordinates")

    def shared_bias(
        self, 
        force_sample,
        sync_interval: int=5,
        mw_file: str="../shared_bias",
        n_trials: int=10,
    ):
        """Syncs ABF bias with buffer file. Excecuted if multiple_walker=True
        
        TODO: 2D collective variables

        Args:
            force_sample: force sample of current step
            sync_interval: number of steps between sychronisation
            mw_file: name of buffer file for shared-bias
            n_trials: number of attempts to access of buffer file before throwing an error
        """
        md_state = self.the_md.get_sampling_data()
        if md_state.step == 0:        
            if self.verbose:
                print(" >>> Info: Creating a new instance for shared-bias ABF.")
                print(" >>> Info: Data of local walker stored in `restart_abf_local.npz`.")
            
            # create seperate restart file with local data only
            self._write_restart(
                filename="restart_abf_local",
                hist=self.histogram,
                force=self.bias,
                m2=self.m2_force,
            )
            
            self.update_samples = np.zeros(shape=(sync_interval, len(force_sample)))
            
            if not os.path.isfile(mw_file+".npz"):
                if self.verbose:
                    print(f" >>> Info: Creating buffer file for shared-bias ABF: `{mw_file}.npz`.")
                self._write_restart(
                    filename=mw_file,
                    hist=self.histogram,
                    force=self.bias,
                    m2=self.m2_force,
                )
                os.chmod(mw_file + ".npz", 0o444)
            elif self.verbose:
                print(f" >>> Info: Syncing with existing buffer file for shared-bias ABF: `{mw_file}.npz`.")
        
        count = md_state.step % sync_interval
        self.update_samples[count] = force_sample
            
        if count == sync_interval-1:
            
            hist = np.zeros_like(self.histogram)
            m2 = np.zeros_like(self.m2_force)
            bias = np.zeros_like(self.bias)

            for sample in self.update_samples:
                bink = (int(sample[1]), int(sample[0]))
                hist[bink] += 1
                for i in range(self.ncoords):
                    (
                        bias[i][bink],
                        m2[i][bink],
                        _,
                    ) = welford_var(
                        hist[bink],
                        bias[i][bink],
                        m2[i][bink],
                        sample[2+i],
                    )              
            # write data of local walker
            self._update_abf(
                "restart_abf_local", 
                hist, bias, m2,
            )             
            
            trial = 0
            while trial < n_trials:
                trial += 1
                if not os.access(mw_file + ".npz", os.W_OK):
                    
                    # grant write access only to local walker during sync
                    os.chmod(mw_file + ".npz", 0o666) 
                    self._update_abf(
                        mw_file, 
                        hist, bias, m2,
                    ) 
                    self.restart(filename=mw_file)
                    os.chmod(mw_file + ".npz", 0o444)                      
                    self.get_pmf()  
                    break                     

                elif trial < n_trials:
                    if self.verbose:
                        print(f" >>> Warning: Retry to open shared buffer file after {trial} failed attempts.")
                    time.sleep(0.1)
                else:
                    raise Exception(f" >>> Fatal Error: Failed to sync bias with `{mw_file}.npz`.")

    def _divergence_xi(self, xi, cv):
        """Calculate divergence of collective variable"""
        if cv.lower() == "distance":
            div = 2.0 / xi
        elif cv.lower() == "angle":
            div = 1.0 / np.tan(xi)
        elif cv.lower() in ["torsion", "2d"]:
            div = 0.0
        else:
            raise NotImplementedError(f" >>> Fatal Error: ABF not implemented for `{cv}`.")

        return div

    def _update_abf(
        self, 
        filename: str, 
        hist: np.ndarray, 
        bias: np.ndarray, 
        m2: np.ndarray,
    ):

        new_hist = np.zeros_like(self.histogram).flatten()
        new_m2 = np.zeros_like(self.m2_force).flatten()
        new_bias = np.zeros_like(self.bias).flatten() 

        with np.load(f"{filename}.npz") as local_data:
            for i in range(len(new_hist)):
                # overwriting global arrays with local data
                (
                    new_hist[i],
                    new_bias[i],
                    new_m2[i],
                    _,
                ) = combine_welford_stats(
                    local_data["hist"].flatten()[i], 
                    local_data["force"].flatten()[i], 
                    local_data["m2"].flatten()[i], 
                    hist.flatten()[i], 
                    bias.flatten()[i], 
                    m2.flatten()[i],
                )
                        
        self._write_restart(
            filename=filename,
            hist=new_hist.reshape(self.histogram.shape),
            force=new_bias.reshape(self.bias.shape),
            m2=new_m2.reshape(self.m2_force.shape),
        )

    def write_restart(self, filename: str = "restart_abf"):
        """write restart file

        Args:
            filename: name of restart file
        """
        self._write_restart(
            filename=filename,
            hist=self.histogram,
            force=self.bias,
            m2=self.m2_force,
        )

    def restart(self, filename: str = "restart_abf"):
        """restart from restart file

        Args:
            filename: name of restart file
        """
        try:
            data = np.load(filename + ".npz")
        except:
            raise OSError(f" >>> Fatal Error: restart file `{filename}.npz` not found!")

        self.histogram = data["hist"]
        self.bias = data["force"]
        self.m2_force = data["m2"]

        if self.verbose:
            print(f" >>> Info: ABF restartet from `{filename}.npz`!")

    def write_traj(self):
        """save trajectory for post-processing"""
        data = {
            "Epot [H]": self.epot,
            "T [K]": self.temp,
        }
        self._write_traj(data)

        # reset trajectories to save memory
        self.traj = np.array([self.traj[-1]])
        self.epot = [self.epot[-1]]
        self.temp = [self.temp[-1]]



import os
import time
import numpy as np

from adaptive_sampling.units import R_in_SI
from .enhanced_sampling import EnhancedSampling


class Reference(EnhancedSampling):
    """Unbiased simulation with restraint to region of interest in cv space with harmonic walls
    Can be used for equilibration prior to production or free energy estimation from unbiased simulation

    Args:
        md: Object of the MD Inteface
        cv_def: definition of the Collective Variable (CV) (see adaptive_sampling.colvars)
                [["cv_type", [atom_indices], minimum, maximum, bin_width], [possible second dimension]]
        equil_temp: equillibrium temperature of MD
        verbose: print verbose information
        kinetice: calculate necessary data to obtain kinetics of reaction
        f_conf: force constant for confinement of system to the range of interest in CV space
        output_freq: frequency in steps for writing outputs
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def step_bias(self, write_output: bool = True, write_traj: bool = True, **kwargs):

        md_state = self.the_md.get_sampling_data()
        (xi, delta_xi) = self.get_cv(**kwargs)

        self.traj = np.append(self.traj, [xi], axis=0)
        self.temp.append(md_state.temp)
        self.epot.append(md_state.epot)

        bias_force = np.zeros_like(md_state.forces)

        if self._check_boundaries(xi):

            bink = self.get_index(xi)
            self.histogram[bink[1], bink[0]] += 1

            if self.kinetics:
                for i in range(self.ncoords):
                    self.cv_crit[i][bink[1], bink[0]] += abs(
                        np.dot(delta_xi[i], md_state.forces)
                    )

        else:
            bias_force = self.harmonic_walls(xi, delta_xi)

        # correction for kinetics
        if self.kinetics:
            self._kinetics(delta_xi)

        if md_state.step % self.out_freq == 0:
            # write output

            if write_traj:
                self.write_traj()

            if write_output:
                self.get_pmf()
                output = {"hist": self.histogram, "free energy": self.pmf}
                self.write_output(output, filename="reference.out")
                self.write_restart()

        return bias_force

    def get_pmf(self) -> np.ndarray:
        """get free energy profile from histogram"""
        self.pmf = (
            -R_in_SI
            / 1000.0  # kJ/mol
            * self.equil_temp
            * np.log(
                self.histogram,
                out=np.zeros_like(self.histogram),
                where=(0 != self.histogram),
            )
        )
        self.pmf -= self.pmf.min()
        return self.pmf

    def shared_bias(
        self, filename: str = "./shared_bias", sync_interval: int = 100, trial: int = 10
    ):
        """sync histogram between multiple walkers

        TODO: fix me

        Args:
            filename: name of bias buffer
            sync_interval: interval in MD steps between syncs of bias
            trial: number of recursive calls if buffer is occupied by other walkers
        """
        # TODO: fix update here
        if os.path.isfile(filename):
            if not os.access(filename, os.W_OK):

                # deny access for other walkers during sync
                os.chmod(filename + ".npz", 0o644)
                sb = np.load(filename + ".npz", allow_pickle=True)
                try:
                    diff = self._last_sync_histogram - self.histogram
                except:
                    if self.verbose:
                        print(" >>> Info: new shared-bias instance created!")
                    self._local_histogram = np.copy(self.histogram)
                    self._last_sync_histogram = np.copy(self.histogram)
                    diff = np.copy(self.histogram)
                os.chmod(filename + ".npz", 0o444)

                self._local_histogram += diff
                self.histogram += diff
                self._last_sync_histogram = np.copy(self.histogram)

            elif trial > 0:
                # recursive calls to wait for other walker to finish writing
                time.sleep(0.1)
                self.shared_bias(filename, sync_interval=sync_interval, trial=trial - 1)

            elif self.verbose:
                print(f" >>> Warning: failed to sync bias with {filename}!")

        else:
            if self.verbose:
                print(f" >>> Info: created shared-bias buffer {filename}!")
            self.write_restart(filename=filename)
            self.write_restart(filename=filename + "_local")
            os.chmod(filename + ".npz", 0o444)

    def write_restart(self, filename: str = "restart_ref"):
        """write restart file

        Args:
            filename: name of restart file
        """
        self._write_restart(filename=filename, hist=self.histogram, pmf=self.pmf)

    def restart(self, filename: str = "restart_ref"):
        """restart from restart file

        Args:
            filename: name of restart file
        """
        try:
            data = np.load(filename + ".npz", allow_pickle=True)
        except:
            raise OSError(f" >>> fatal error: restart file {filename}.npz not found!")

        self.histogram = data["hist"]
        self.pmf = data["pmf"]

        if self.verbose:
            print(f" >>> Info: Sampling restartet from {filename}!")

    def write_traj(self):
        """save trajectory for post-processing"""
        data = {
            "Epot [H]": self.epot,
            "T [K]": self.temp,
        }
        self._write_traj(data)

        # reset trajectories to save memory
        self.traj = np.array([self.traj[-1]])
        self.epot = [self.epot[-1]]
        self.temp = [self.temp[-1]]



from fileinput import filename
import os, time, itertools
import numpy as np
from .enhanced_sampling import EnhancedSampling
from .utils import welford_var, combine_welford_stats, diff
from .eabf import eABF
from .metadynamics import WTM
from ..units import *


class WTMeABF(eABF, WTM, EnhancedSampling):
    """Well-Tempered Metadynamics extended-system Adaptive Biasing Force method

       see: Fu et. al., J. Phys. Chem. Lett. (2018); https://doi.org/10.1021/acs.jpclett.8b01994

    The collective variable is coupled to a fictitious particle with an harmonic force.
    The dynamics of the fictitious particel is biased using a combination of ABF and Metadynamics.

    Args:
        ext_sigma: thermal width of coupling between collective and extended variable
        ext_mass: mass of extended variable in atomic units
        hill_height: height of Gaussian hills in kJ/mol
        hill_std: standard deviation of Gaussian hills in units of the CV (can be Bohr, Degree, or None)
        md: Object of the MD Inteface
        cv_def: definition of the Collective Variable (CV) (see adaptive_sampling.colvars)
                [["cv_type", [atom_indices], minimum, maximum, bin_width], [possible second dimension]]
        friction: friction coefficient for Lagevin dynamics of the extended-system
        seed_in: random seed for Langevin dynamics of extended-system
        nfull: Number of force samples per bin where full bias is applied,
               if nsamples < nfull the bias force is scaled down by nsamples/nfull
        hill_drop_freq: frequency of hill creation in steps
        well_tempered_temp: effective temperature for WTM, if None, hills are not scaled down (normal metadynamics)
        force_from_grid: forces are accumulated on grid for performance,
                         if False, forces are calculated from sum of Gaussians in every step
        equil_temp: equillibrium temperature of MD
        verbose: print verbose information
        kinetice: calculate necessary data to obtain kinetics of reaction
        f_conf: force constant for confinement of system to the range of interest in CV space
        output_freq: frequency in steps for writing outputs

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.abf_forces = np.zeros_like(self.bias)

    def step_bias(
        self, 
        write_output: bool = True, 
        write_traj: bool = True, 
        output_file: str = 'wtmeabf.out',
        traj_file: str = 'CV_traj.dat', 
        restart_file: str = 'restart_wtmeabf',
        **kwargs):

        md_state = self.the_md.get_sampling_data()
        (xi, delta_xi) = self.get_cv(**kwargs)
        self._propagate()

        mtd_forces = self.get_wtm_force(self.ext_coords)
        bias_force = self._extended_dynamics(xi, delta_xi)  # , self.hill_std)
        force_sample = [0 for _ in range(2 * self.ncoords)]

        if (self.ext_coords <= self.maxx).all() and (
            self.ext_coords >= self.minx
        ).all():

            bin_la = self.get_index(self.ext_coords)
            self.ext_hist[bin_la[1], bin_la[0]] += 1

            for i in range(self.ncoords):

                # linear ramp function
                ramp = (
                    1.0
                    if self.ext_hist[bin_la[1], bin_la[0]] > self.nfull
                    else self.ext_hist[bin_la[1], bin_la[0]] / self.nfull
                )

                # apply bias force on extended variable
                force_sample[i] = self.ext_k[i] * diff(self.ext_coords[i], xi[i], self.cv_type[i])
                (
                    self.abf_forces[i][bin_la[1], bin_la[0]],
                    self.m2_force[i][bin_la[1], bin_la[0]],
                    self.var_force[i][bin_la[1], bin_la[0]],
                ) = welford_var(
                    self.ext_hist[bin_la[1], bin_la[0]],
                    self.abf_forces[i][bin_la[1], bin_la[0]],
                    self.m2_force[i][bin_la[1], bin_la[0]],
                    force_sample[i],
                )
                self.ext_forces -= (
                    ramp * self.abf_forces[i][bin_la[1], bin_la[0]] - mtd_forces[i]
                )

        # xi-conditioned accumulators for CZAR
        if self._check_boundaries(xi):
            bink = self.get_index(xi)
            self.histogram[bink[1], bink[0]] += 1

            for i in range(self.ncoords):
                force_sample[self.ncoords+i] = self.ext_k[i] * diff(
                    self.ext_coords[i], self.grid[i][bink[i]], self.cv_type[i]
                )
                self.correction_czar[i][bink[1], bink[0]] += force_sample[self.ncoords+i]

        # shared-bias eABF
        if self.shared:                        
            self.shared_bias(
                xi,
                force_sample, 
                **kwargs,
            )

        self.traj = np.append(self.traj, [xi], axis=0)
        self.ext_traj = np.append(self.ext_traj, [self.ext_coords], axis=0)
        self.temp.append(md_state.temp)
        self.epot.append(md_state.epot)
        self._up_momenta()

        # correction for kinetics
        if self.kinetics:
            self._kinetics(delta_xi)

        if md_state.step % self.out_freq == 0:
            # write output

            if write_traj:
                self.write_traj(filename=traj_file)

            if write_output:
                self.get_pmf()
                output = {"hist": self.histogram, "free energy": self.pmf}
                for i in range(self.ncoords):
                    output[f"metaforce {i}"] = self.bias[i]
                    output[f"abf force {i}"] = self.abf_forces[i]
                    output[f"czar force {i}"] = self.czar_force[i]
                    # TODO: output variance of CZAR for error estimate
                    # output[f"var force {i}"] = self.var_force[i]
                output[f"metapot"] = self.metapot

                self.write_output(output, filename=output_file)
                self.write_restart(filename=restart_file)

        return bias_force

    def shared_bias(
        self, 
        xi,
        force_sample,
        sync_interval: int=50,
        mw_file: str="../shared_bias",
        local_file: str="restart_wtmeabf_local",
        n_trials: int=10,
    ):
        """syncs eABF bias with other walkers

        TODO: 2D collective variables

        Args:
            xi: state of the collective variable
            force_sample: force sample of current step
            sync_interval: number of steps between sychronisation
            mw_file: name of buffer file for shared-bias
            n_trials: number of attempts to access of buffer file before throwing an error
        """
        md_state = self.the_md.get_sampling_data()
        if md_state.step == 0:        
            if self.verbose:
                print(" >>> Info: Creating a new instance for shared-bias eABF.")
                print(f" >>> Info: Data of local walker stored in `{local_file}.npz`.")
            
            # create seperate restart file with local data only
            self._write_restart(
                filename=local_file,
                hist=self.histogram,
                force=self.bias,
                m2=self.m2_force,
                ext_hist=self.ext_hist,
                czar_corr=self.correction_czar,
                abf_force=self.abf_forces,
                center=self.center,
                metapot=self.metapot,
            )
            
            if sync_interval % self.hill_drop_freq != 0:
                raise ValueError(
                    " >>> Fatal Error: Sync interval for shared-bias WTM has to divisible through the frequency of hill creation!"
                )

            self.len_center_last_sync = len(self.center)
            self.update_samples = np.zeros(shape=(sync_interval, len(force_sample)))
            self.last_samples_xi = np.zeros(shape=(sync_interval, len(xi)))
            self.last_samples_la = np.zeros(shape=(sync_interval, len(self.ext_coords)))
            self.metapot_last_sync = np.copy(self.metapot)
            self.bias_last_sync = np.copy(self.bias)

            if not os.path.isfile(mw_file+".npz"):
                if self.verbose:
                    print(f" >>> Info: Creating buffer file for shared-bias WTM-eABF: `{mw_file}.npz`.")
                self._write_restart(
                    filename=mw_file,
                    hist=self.histogram,
                    force=self.bias,
                    m2=self.m2_force,
                    ext_hist=self.ext_hist,
                    czar_corr=self.correction_czar,
                    abf_force=self.abf_forces,
                    center=self.center,
                    metapot=self.metapot,
                )
                os.chmod(mw_file + ".npz", 0o444)
            elif self.verbose:
                print(f" >>> Info: Syncing with buffer file `{mw_file}.npz`.")
        
        # save new samples
        count = md_state.step % sync_interval
        self.update_samples[count] = force_sample
        self.last_samples_xi[count] = xi
        self.last_samples_la[count] = self.ext_coords

        if count == sync_interval-1:
            
            # calculate progress since last sync from new samples
            hist = np.zeros_like(self.histogram)
            m2 = np.zeros_like(self.m2_force)
            abf_forces = np.zeros_like(self.abf_forces)
            ext_hist = np.zeros_like(self.ext_hist)
            czar_corr = np.zeros_like(self.correction_czar)
            
            delta_bias = self.bias - self.bias_last_sync
            delta_metapot = self.metapot - self.metapot_last_sync
            new_center = self.center[self.len_center_last_sync:]            

            for i, sample in enumerate(self.update_samples):
                if self._check_boundaries(self.last_samples_la[i]):                   
                    bin_la = self.get_index(self.last_samples_la[i])
                    ext_hist[bin_la[1], bin_la[0]] += 1
                    for j in range(self.ncoords):
                        (
                            abf_forces[j][bin_la[1], bin_la[0]],
                            m2[j][bin_la[1], bin_la[0]],
                            _,
                        ) = welford_var(
                            ext_hist[bin_la[1], bin_la[0]],
                            abf_forces[j][bin_la[1], bin_la[0]],
                            m2[j][bin_la[1], bin_la[0]],
                            sample[j],
                        )
                
                if self._check_boundaries(self.last_samples_xi[i]):
                    bin_xi = self.get_index(self.last_samples_xi[i])
                    hist[bin_xi[1], bin_xi[0]] += 1 
                    for j in range(self.ncoords):
                        czar_corr[j][bin_xi[1], bin_xi[0]] += sample[self.ncoords+j]
            
            # add new samples to local restart
            self._update_wtmeabf(
                local_file,
                hist, 
                ext_hist,
                abf_forces,
                m2,
                czar_corr,
                delta_bias,
                delta_metapot,
                new_center,
            )

            trial = 0
            while trial < n_trials:
                trial += 1
                if not os.access(mw_file + ".npz", os.W_OK):
                    
                    # grant write access only to one walker during sync
                    os.chmod(mw_file + ".npz", 0o666) 
                    self._update_wtmeabf(
                        mw_file,
                        hist, 
                        ext_hist,
                        abf_forces,
                        m2,
                        czar_corr,
                        delta_bias,
                        delta_metapot,
                        new_center,
                    )
                    self.restart(filename=mw_file)
                    os.chmod(mw_file + ".npz", 0o444)  # other walkers can access again

                    self.get_pmf()  # get new global pmf
                    self.metapot_last_sync = np.copy(self.metapot)
                    self.bias_last_sync = np.copy(self.bias)
                    self.len_center_last_sync = len(self.center)
                    break                     

                elif trial < n_trials:
                    if self.verbose:
                        print(f" >>> Warning: Retry to open shared buffer file after {trial} failed attempts.")
                    time.sleep(0.1)
                else:
                    raise Exception(f" >>> Fatal Error: Failed to sync bias with `{mw_file}.npz`.")

    def _update_wtmeabf(
        self, 
        filename: str,
        hist: np.ndarray, 
        ext_hist: np.ndarray,
        abf_forces: np.ndarray,
        m2: np.ndarray,
        czar_corr: np.ndarray,
        delta_bias: np.ndarray,
        delta_metapot: np.ndarray,
        center: np.ndarray,
    ):
        with np.load(f"{filename}.npz") as data:

            new_hist = data["hist"] + hist      
            new_czar_corr = data["czar_corr"] + czar_corr
            new_metapot = data["metapot"] + delta_metapot
            new_bias = data["force"] + delta_bias
            
            new_m2 = np.zeros_like(self.m2_force).flatten()
            new_abf_forces = np.zeros_like(self.abf_forces).flatten()
            new_ext_hist = np.zeros_like(self.ext_hist).flatten()
            new_centers = np.append(data["center"], center), 
            
            for i in range(len(hist.flatten())):
                (
                    new_ext_hist[i],
                    new_abf_forces[i],
                    new_m2[i],
                    _,
                ) = combine_welford_stats(
                    data["ext_hist"].flatten()[i], 
                    data["abf_force"].flatten()[i], 
                    data["m2"].flatten()[i], 
                    ext_hist.flatten()[i], 
                    abf_forces.flatten()[i], 
                    m2.flatten()[i], 
                )
                                    
        self._write_restart(
            filename=filename,
            hist=new_hist,
            force=new_bias,
            m2=new_m2.reshape(self.m2_force.shape),
            ext_hist=new_ext_hist.reshape(self.histogram.shape),
            czar_corr=new_czar_corr,
            abf_force=new_abf_forces.reshape(self.abf_forces.shape),
            center=np.asarray(list(itertools.chain(*new_centers))),
            metapot=new_metapot,
        )                           

    def write_restart(self, filename: str = "restart_wtmeabf"):
        """write restart file

        Args:
            filename: name of restart file
        """
        self._write_restart(
            filename=filename,
            hist=self.histogram,
            force=self.bias,
            m2=self.m2_force,
            ext_hist=self.ext_hist,
            czar_corr=self.correction_czar,
            abf_force=self.abf_forces,
            center=self.center,
            metapot=self.metapot,
        )

    def restart(self, filename: str = "restart_wtmeabf"):
        """restart from restart file

        Args:
            filename: name of restart file
        """
        try:
            data = np.load(filename + ".npz", allow_pickle=True)
        except:
            raise OSError(f" >>> fatal error: restart file {filename}.npz not found!")

        self.histogram = data["hist"]
        self.bias = data["force"]
        self.m2_force = data["m2"]
        self.ext_hist = data["ext_hist"]
        self.correction_czar = data["czar_corr"]
        self.abf_forces = data["abf_force"]
        self.center = data["center"].tolist()
        self.metapot = data["metapot"]

        if self.verbose:
            print(f" >>> Info: Adaptive sampling restartet from {filename}!")

    def write_traj(self, filename: str = 'CV_traj.dat'):
        """save trajectory for post-processing"""

        data = self._write_ext_traj()
        data["Epot [H]"] = self.epot
        data["T [K]"] = self.temp

        self._write_traj(data, filename=filename)

        # reset trajectories to save memory
        self.traj = np.array([self.traj[-1]])
        self.ext_traj = np.array([self.ext_traj[-1]])
        self.epot = [self.epot[-1]]
        self.temp = [self.temp[-1]]



import sys
import numpy as np
from typing import List, Tuple
from .utils import join_frames
from ..units import *


def run_mbar(
    exp_U: np.ndarray,
    frames_per_traj: np.ndarray, 
    max_iter: int = 10000,
    conv: float = 1.0e-7,
    conv_errvec: float = None,
    outfreq: int = 100,
) -> np.ndarray:
    """Self-consistent Multistate Bannett Acceptance Ratio (MBAR)

       see: Shirts et. al., J. Chem. Phys. (2008); https://doi.org/10.1063/1.2978177

    Args:
        exp_U: num_trajs*num_frames array of biased Boltzman factors
        frames_per_traj: number of samples per trajectory
        max_iter: Maximum number of iterations
        conv: Convergence criterion, largest change in beta*Ai
        conv_errvec: Convergence criterion based on the error vector, not used if None, largest absolute value in the error vec
        outfreq: Output frequency during self-consistent iteration

    Returns:
        W: array containing statistical weigths of each frame
    """
    num_trajs = len(frames_per_traj)
    beta_Ai = np.zeros(shape=(num_trajs,), dtype=float)
    Ai_overtime = [beta_Ai]

    print("All ready!\n")
    print("Start of the self-consistent iteration.")
    print("========================================================================")
    sys.stdout.flush()
    count = 0
    while True:
        count += 1

        denominator = np.multiply(frames_per_traj * np.exp(beta_Ai), exp_U.T)
        denominator = 1.0 / np.sum(denominator, axis=1)

        beta_Ai_new = -np.log(np.multiply(exp_U, denominator).sum(axis=1))
        beta_Ai_new -= beta_Ai_new[0]
        Ai_overtime.append(beta_Ai_new)

        delta_Ai = np.abs(beta_Ai - beta_Ai_new)
        beta_Ai = np.copy(beta_Ai_new)

        max_err_vec = np.abs(_error_vec(frames_per_traj, beta_Ai, exp_U)).max()

        if count % outfreq == 0 or count == 1:
            print(
                "Iter %4d:\tConv=%14.10f\tConv_errvec=%14.6f"
                % (count, np.max(delta_Ai[1:]), max_err_vec)
            )
            sys.stdout.flush()

        if conv_errvec == None:
            converged = delta_Ai[1:].max() < conv
        else:
            converged = (delta_Ai[1:].max() < conv) and (max_err_vec < conv_errvec)

        if count == max_iter:
            print(
                "========================================================================"
            )
            print(f"Convergence not reached in {count} iterations!")
            print("Max error vector:", max_err_vec)
            print(
                "========================================================================"
            )
            sys.stdout.flush()
            break

        if converged:
            print(
                "========================================================================"
            )
            print(f"Converged after {count} iterations!")
            print("Max error vector:", max_err_vec)
            print(
                "========================================================================"
            )
            sys.stdout.flush()
            break

    # final values
    weights = np.multiply(frames_per_traj * np.exp(beta_Ai), exp_U.T)
    weights = 1.0 / np.sum(weights, axis=1)

    return weights

def build_boltzmann(
    traj_list: list, 
    meta_f: np.ndarray, 
    dU_list: list=None, 
    equil_temp: float=300.0
) -> Tuple[np.ndarray, np.ndarray]:
    """Build Boltzmann factores for MBAR
    
    Args:
        traj_list: list of trajectories
        meta_f: definition of simulation with force constants and cv grid
        dU_list: optional, additional potential that enters boltzmann factor (GaMD, confinement, reweighting, ...)
    
    Returns:
        exp_U: num_trajs**num_frames array of Boltzmann factors
        frames_per_traj: Number of frames per trajectory
    """
    RT = R_in_SI * equil_temp / 1000.0
    beta = 1.0 / RT
    
    all_frames, num_frames, frames_per_traj = join_frames(traj_list)
    if dU_list is not None:
        all_dU, dU_num, dU_per_traj = join_frames(dU_list)
        all_dU *= atomic_to_kJmol  # kJ/mol
        if (dU_num != num_frames) or (frames_per_traj != dU_per_traj).all():
            raise ValueError(" >>> Error: GaMD frames have to match eABF frames!")

    exp_U = []
    for _, line in enumerate(meta_f):
        if dU_list:
            exp_U.append(
                np.exp(
                    -beta
                    * (all_dU + 0.5 * line[2] * np.power(all_frames - line[1], 2)),
                    dtype=float,
                )
            )
        else:
            exp_U.append(
                np.exp(
                    -beta * 0.5 * line[2] * np.power(all_frames - line[1], 2),
                    dtype=float,
                )
            )

    exp_U = np.asarray(exp_U, dtype=float)  # this is a num_trajs x num_frames array
    return exp_U, frames_per_traj

def get_windows(
    centers: np.ndarray,
    xi: np.ndarray,
    la: np.ndarray,
    sigma: float,
    equil_temp: float = 300.0,
) -> Tuple[List[np.ndarray], np.ndarray, np.ndarray]:
    """generate mixture distribution of Gaussian shaped windows from eABF trajectory
       
       see: J. Chem. Phys. 157, 024110 (2022); https://doi.org/10.1063/5.0095554

    Args:
        centers: centers of windows
        xi: Trajectory of the reaction coordinate
        la: Trajectory of the extended variable
        sigma: Thermal width of coupling of xi and la
        equil_temp: equillibrium temperature

    Returns:
        traj_list: list of window trajectories,
        index_list: list of frame indices in original trajectory,
        meta_f: window information for MBAR (compatible with other popular MBAR/WHAM implementations)
    """
    RT = R_in_SI * equil_temp / 1000.0
    k = RT / (sigma * sigma)

    dx = centers[1] - centers[0]
    dx2 = dx / 2.0

    traj_list = []
    index_list = np.array([])
    for center in centers:
        indices = np.where(np.logical_and(la >= center - dx2, la < center + dx2))
        index_list = np.append(index_list, indices[0])
        traj_list += [xi[indices]]

    meta_f = np.zeros(shape=(len(centers), 3))
    meta_f[:, 1] = centers
    meta_f[:, 2] = k

    return traj_list, index_list.astype(np.int32), meta_f


def pmf_from_weights(
    grid: np.array, cv: np.array, weights: np.array, equil_temp: float = 300.0
) -> Tuple[np.array, np.array]:
    """make free energy surface from statistical weigths obtained by MBAR

    Args:
        grid: grid along cv
        cv: trajectory of cv
        weights: boltzmann weights of frames in trajectory
        equil_temp: Temperature of simulation

    Returns:
        Potential of mean force (PMF), probability density
    """
    RT = R_in_SI * equil_temp / 1000.0

    dx = grid[1] - grid[0]
    dx2 = dx / 2.0

    rho = np.zeros(len(grid), dtype=float)
    for ii, x in enumerate(grid):
        W_x = weights[np.where(np.logical_and(cv >= x - dx2, cv < x + dx2))]
        rho[ii] = W_x.sum()

    rho /= rho.sum() * dx
    pmf = -RT * np.log(rho, out=np.zeros_like(rho), where=(rho != 0))

    return pmf, rho


def deltaf_from_weights(
    TS: float, cv: np.array, weights: np.array, equil_temp: float = 300.0
) -> Tuple[np.array, np.array]:
    """Compute free energy difference from statistical weigths obtained by MBAR

    Args:
        TS: position of transition state on CV
        cv: trajectory of CV
        weights: MBAR weights of data points
        equil_temp: temperature

    Returns:
        deltaF: free energy difference
    """
    RT = R_in_SI * equil_temp / 1000.0
    weights /= weights.sum()

    p_a = weights[np.where(cv < TS)].sum()
    p_b = weights[np.where(cv > TS)].sum()

    return -RT * np.log(p_b / p_a)


def _error_vec(
    n_frames: np.ndarray, beta_Ai: np.ndarray, exp_U: np.ndarray
) -> np.ndarray:
    """error vector for MBAR"""
    denominator = np.multiply(n_frames * np.exp(beta_Ai), exp_U.T)
    denominator = 1.0 / np.sum(denominator, axis=1)
    # sum over all frames
    error_v = n_frames - n_frames * np.exp(beta_Ai) * (
        np.multiply(exp_U, denominator).sum(axis=1)
    )
    return error_v



import os, time, itertools
import numpy as np
from typing import Tuple
from .enhanced_sampling import EnhancedSampling
from .utils import diff, sum
from ..units import *
from ..processing_tools.thermodynamic_integration import integrate


class WTM(EnhancedSampling):
    """Well-Tempered Metadynamics

       see: Barducci et. al., Phys. rev. lett. (2008); https://doi.org/10.1103/PhysRevLett.100.020603

    An repulsive biasing potential is built by a superposition of Gaussian hills along the reaction coordinate.

    Args:
        hill_height: height of Gaussian hills in kJ/mol
        hill_std: standard deviation of Gaussian hills in units of the CV (can be Bohr, Degree, or None)
        hill_drop_freq: frequency of hill creation in steps
        well_tempered_temp: effective temperature for WTM, if None, hills are not scaled down (normal metadynamics)
        force_from_grid: forces are accumulated on grid for performance (recommended),
                         if False, forces are calculated from sum of Gaussians in every step
        estimator: if "TI", PMF is estimated from integral of bias force, else PMF directly estimated from force
    """

    def __init__(
        self,
        hill_height: float,
        hill_std: list,
        *args,
        hill_drop_freq: int = 20,
        well_tempered_temp: float = 3000.0,
        force_from_grid: bool = True,
        estimator: str = "TI",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if int(hill_drop_freq) <= 0:
            raise ValueError(" >>> Error: Update interval has to be int > 0!")

        if hill_height <= 0:
            raise ValueError(" >>> Error: Gaussian height for MtD has to be > 0!")

        if well_tempered_temp is None and self.verbose:
            print(" >>> Info: well-tempered scaling of WTM hill_height switched off")
        elif well_tempered_temp <= 0:
            raise ValueError(
                " >>> Error: Effective temperature for Well-Tempered MtD has to be > 0!"
            )

        hill_std = [hill_std] if not hasattr(hill_std, "__len__") else hill_std

        self.hill_height = hill_height / atomic_to_kJmol
        self.hill_std = self.unit_conversion_cv(np.asarray(hill_std))[0]
        self.hill_var = self.hill_std * self.hill_std
        self.hill_drop_freq = int(hill_drop_freq)
        self.well_tempered_temp = well_tempered_temp
        self.well_tempered = False if well_tempered_temp == None else True
        self.force_from_grid = force_from_grid
        self.estimator = estimator

        self.metapot = np.zeros_like(self.histogram)
        self.center = []

    def step_bias(self, write_output: bool = True, write_traj: bool = True, **kwargs):

        md_state = self.the_md.get_sampling_data()
        (xi, delta_xi) = self.get_cv(**kwargs)

        bias_force = np.zeros_like(md_state.forces)

        mtd_force = self.get_wtm_force(xi)
        if self._check_boundaries(xi):

            bink = self.get_index(xi)
            self.histogram[bink[1], bink[0]] += 1

            for i in range(self.ncoords):
                bias_force += mtd_force[i] * delta_xi[i]

        bias_force += self.harmonic_walls(xi, delta_xi)  # , 1.6 * self.hill_std)

        self.traj = np.append(self.traj, [xi], axis=0)
        self.temp.append(md_state.temp)
        self.epot.append(md_state.epot)

        # correction for kinetics
        if self.kinetics:
            self._kinetics(delta_xi)

        # shared-bias metadynamics
        if self.shared:
            self.shared_bias(xi, **kwargs)

        if self.the_md.step % self.out_freq == 0:
            # write output

            if write_traj:
                self.write_traj()

            if write_output:
                self.get_pmf()
                output = {
                    "hist": self.histogram,
                    "free energy": self.pmf,
                    "metapot": self.metapot * atomic_to_kJmol,
                }

                self.write_output(output, filename="wtm.out")
                self.write_restart()

        return bias_force

    def get_pmf(self):
        if self.estimator == "TI" and self.ncoords == 1:
            # PMF from integration of bias force
            self.pmf[0, :], _ = integrate(
                -self.bias[0][0],
                self.dx,
                equil_temp=self.equil_temp,
                method="trapezoid",
            )

        else:
            # PMF from bias potential
            self.pmf = -self.metapot

        if self.well_tempered:
            self.pmf *= (
                self.equil_temp + self.well_tempered_temp
            ) / self.well_tempered_temp
        self.pmf *= atomic_to_kJmol
        self.pmf -= self.pmf.min()

    def shared_bias(
        self,
        xi, 
        sync_interval: int=50,
        mw_file: str="shared_bias",
        n_trials: int=10,
    ):
        """TODO"""
        md_state = self.the_md.get_sampling_data()
        if md_state.step == 0:        
            if self.verbose:
                print(" >>> Info: Creating a new instance for shared-bias metadynamics.")
                print(" >>> Info: Data of local walker stored in `restart_wtm_local.npz`.")
            
            # create seperate restart file with local data only
            self._write_restart(
                filename="restart_wtm_local",
                hist=self.histogram,
                force=self.bias,
                metapot=self.metapot,
                centers=self.center,
            )
            
            if sync_interval % self.hill_drop_freq != 0:
                raise ValueError(
                    " >>> Fatal Error: Sync interval for shared-bias WTM has to divisible through the frequency of hill creation!"
                )

            self.len_center_last_sync = len(self.center)
            self.metapot_last_sync = np.copy(self.metapot)
            self.bias_last_sync = np.copy(self.bias)
            self.traj_last_sync = np.zeros(shape=(sync_interval, len(xi)))
            if not os.path.isfile(mw_file+".npz"):
                if self.verbose:
                    print(f" >>> Info: Creating buffer file for shared-bias metadynamics: `{mw_file}.npz`.")
                self._write_restart(
                    filename=mw_file,
                    hist=self.histogram,
                    force=self.bias,
                    metapot=self.metapot,
                    centers=self.center,
                )
                os.chmod(mw_file + ".npz", 0o444)
            elif self.verbose:
                print(f" >>> Info: Syncing with existing buffer file for shared-bias metadynamics: `{mw_file}.npz`.")
        
        count = md_state.step % sync_interval
        self.traj_last_sync[count] = xi
        if count == sync_interval-1:
            
            new_center = self.center[self.len_center_last_sync:]            
            delta_hist = np.zeros_like(self.histogram)
            for x in self.traj_last_sync:
                if self._check_boundaries(x):
                    bink = self.get_index(x)
                    delta_hist[bink[1], bink[0]] += 1

            delta_bias = self.bias - self.bias_last_sync 
            delta_metapot = self.metapot - self.metapot_last_sync 
                        
            self._update_wtm(
                "restart_wtm_local",
                delta_hist,
                delta_bias,
                delta_metapot,
                new_center,
            )

            trial = 0
            while trial < n_trials:
                trial += 1
                if not os.access(mw_file + ".npz", os.W_OK):
                    
                    # grant write access only to one walker during sync
                    os.chmod(mw_file + ".npz", 0o666) 
                    self._update_wtm(
                        mw_file,
                        delta_hist,
                        delta_bias,
                        delta_metapot,
                        new_center,
                    )
                    self.restart(filename=mw_file)
                    os.chmod(mw_file + ".npz", 0o444)  # other walkers can access again

                    self.metapot_last_sync = np.copy(self.metapot)
                    self.bias_last_sync = np.copy(self.bias)
                    self.len_center_last_sync = len(self.center)
                    break                     

                elif trial < n_trials:
                    if self.verbose:
                        print(f" >>> Warning: Retry to open shared buffer file after {trial} failed attempts.")
                    time.sleep(0.1)
                else:
                    raise Exception(f" >>> Fatal Error: Failed to sync bias with `{mw_file}.npz`.")   

    def get_wtm_force(self, xi: np.ndarray) -> list:
        """compute well-tempered metadynamics bias force from superposition of gaussian hills

        Args:
            xi: state of collective variable

        Returns:
            bias_force: bias force from metadynamics
        """
        if self.the_md.step % self.hill_drop_freq == 0:
            self.center.append(xi[0]) if self.ncoords == 1 else self.center.append(xi)
            
        is_in_bounds = (xi <= self.maxx).all() and (xi >= self.minx).all()
        bias_force = [0, 0]
        if is_in_bounds:
            bias_force = self._accumulate_wtm_force(xi)

        if not is_in_bounds or not self.force_from_grid:
            bias_force, _ = self._analytic_wtm_force(xi)

        return bias_force

    def _accumulate_wtm_force(self, xi: np.ndarray) -> list:
        """accumulate metadynamics potential and its gradient on grid

        Args:
            xi: state of collective variable

        Returns:
            bias_force: bias force from metadynamics
        """
        if self._check_boundaries(xi):
            bink = self.get_index(xi)
            if self.the_md.step % self.hill_drop_freq == 0:
                if self.ncoords == 1:

                    if self.well_tempered:
                        w = self.hill_height * np.exp(
                            -self.metapot[bink[1], bink[0]]
                            / (kB_in_atomic * self.well_tempered_temp)
                        )
                    else:
                        w = self.hill_height

                    dx = diff(self.grid[0], xi[0], self.cv_type[0])
                    epot = w * np.exp(-(dx * dx) / (2.0 * self.hill_var[0]))
                    self.metapot[0] += epot
                    self.bias[0][0] -= epot * dx / self.hill_var[0]

                    self._smooth_boundary_grid(xi, w)

                else:
                    # TODO: implement for 2D
                    raise NotImplementedError(
                        " >>> Error: metadynamics currently only supported for 1D coordinates"
                    )
            return [self.bias[i][bink[1], bink[0]] for i in range(self.ncoords)]
        
        else:
            return [0.0 for _ in range(self.ncoords)]

    def _analytic_wtm_force(self, xi: np.ndarray) -> Tuple[list, float]:
        """compute analytic WTM bias force from sum of gaussians hills

        Args:
            xi: state of collective variable

        Returns:
            bias_force: bias force from metadynamics
        """
        local_pot = 0.0
        bias_force = [0.0 for _ in range(self.ncoords)]

        if len(self.center) == 0:
            if self.verbose:
                print(" >>> Warning: no metadynamics hills stored")
            return bias_force

        if self.ncoords == 1:

            dx = diff(xi[0], np.asarray(self.center), self.cv_type[0])
            ind = np.ma.indices((len(self.center),))[0]
            ind = np.ma.masked_array(ind)
            ind[abs(dx) > 3 * self.hill_std[0]] = np.ma.masked

            # can get slow in long run, so only iterate over significant elements
            for i in np.nditer(ind.compressed(), flags=["zerosize_ok"]):

                if self.well_tempered:
                    w = self.hill_height * np.exp(
                        -local_pot / (kB_in_atomic * self.well_tempered_temp)
                    )
                else:
                    w = self.hill_height

                epot = w * np.exp(-(dx[i] * dx[i]) / (2.0 * self.hill_var[0]))
                local_pot += epot
                bias_force[0] -= epot * dx[i] / self.hill_var[0]

                local_pot, bias_force = self._smooth_boundary_analytic(
                    xi, w, self.center[i], local_pot, bias_force
                )
        else:
            # TODO: implement for 2D
            pass

        return bias_force, local_pot

    def _smooth_boundary_grid(self, xi: np.ndarray, w: float):
        """ensures linear profile of bias potential at boundary

        see: Crespo et al. (2010), https://doi.org/10.1103/PhysRevE.81.055701

        Args:
            xi: Collective variable
        """
        chi1 = 3.0 * self.hill_std

        if self.ncoords == 1:
            dx_list = []
            w_list = []

            if diff(self.maxx[0], chi1[0], self.cv_type[0]) <= xi[0] <= self.maxx[0]:
                center_out = self.maxx[0] + diff(self.maxx[0], xi[0], self.cv_type[0])
                dx_list.append(diff(self.grid[0], center_out, self.cv_type[0]))
                w_list.append(w)
            else:
                # TODO: scale height of hills to ensure flat potential at boundary
                pass

            if self.minx[0] <= xi[0] <= sum(self.minx[0], chi1[0], self.cv_type[0]):
                center_out = self.minx[0] - diff(xi[0], self.minx[0], self.cv_type[0])
                dx_list.append(diff(self.grid[0], center_out, self.cv_type[0]))
                w_list.append(w)
            else:
                # TODO: scale height of hills to ensure flat potential at boundary
                pass

            for dx, W in zip(dx_list, w_list):
                epot = W * np.exp(-(dx * dx) / (2.0 * self.hill_var[0]))
                self.metapot[0] += epot
                self.bias[0][0] -= epot * dx / self.hill_var[0]

        else:
            # TODO: implement for 2D
            pass

    def _smooth_boundary_analytic(self, xi, w, center, local_pot, bias_force):
        """ensures linear profile at boundary

        see: Crespo et al. (2010), https://doi.org/10.1103/PhysRevE.81.055701

        Args:
            xi: Collective variable
        """
        chi1 = 3.0 * self.hill_std

        if self.ncoords == 1:
            if diff(self.maxx[0], chi1[0], self.cv_type[0]) <= center <= self.maxx[0]:
                center_out = self.maxx[0] + diff(self.maxx[0], center, self.cv_type[0])
                val = diff(xi[0], center_out, self.cv_type[0])
                epot = w * np.exp(-(val * val) / (2.0 * self.hill_var[0]))
                local_pot += epot
                bias_force[0] -= epot * val / self.hill_var[0]
            else:
                # TODO: scale height of hills to ensure flat potential at boundary
                pass

            if self.minx[0] <= center <= sum(self.minx[0], chi1[0], self.cv_type[0]):
                center_out = self.minx[0] - diff(center, self.minx[0], self.cv_type[0])
                val = diff(xi[0], center_out, self.cv_type[0])
                epot = w * np.exp(-(val * val) / (2.0 * self.hill_var[0]))
                local_pot += epot
                bias_force[0] -= epot * val / self.hill_var[0]
            else:
                # TODO: scale height of hills to ensure flat potential at boundary
                pass

        return local_pot, bias_force

    def _update_wtm(
        self,
        filename,
        hist,
        bias,
        metapot,
        center
    ):
        with np.load(f"{filename}.npz") as data:  
            new_hist = data["hist"] + hist
            new_bias = data["force"] + bias
            new_metapot = data["metapot"] + metapot
            new_centers = np.append(data["centers"], center), 
        
        self._write_restart(
            filename=filename,
            hist=new_hist,
            force=new_bias,
            metapot=new_metapot,
            centers=np.asarray(list(itertools.chain(*new_centers))),
        )
                    
    def write_restart(self, filename: str = "restart_wtm"):
        """write restart file

        Args:
            filename: name of restart file
        """
        self._write_restart(
            filename=filename,
            hist=self.histogram,
            force=self.bias,
            metapot=self.metapot,
            centers=self.center,
        )

    def restart(self, filename: str = "restart_wtm"):
        """restart from restart file

        Args:
            filename: name of restart file
        """
        try:
            data = np.load(filename + ".npz", allow_pickle=True)
        except:
            raise OSError(f" >>> fatal error: restart file `{filename}.npz` not found!")

        self.histogram = data["hist"]
        self.bias = data["force"]
        self.metapot = data["metapot"]
        self.center = data["centers"].tolist()

        if self.verbose:
            print(f" >>> Info: Adaptive sampling restartet from `{filename}.npz`!")

    def write_traj(self):
        """save trajectory for post-processing"""
        data = {
            "Epot [H]": self.epot,
            "T [K]": self.temp,
        }
        self._write_traj(data)

        # reset trajectories to save memory
        self.traj = np.array([self.traj[-1]])
        self.epot = [self.epot[-1]]
        self.temp = [self.temp[-1]]



import numpy as np
from ..units import *


def integrate(
    mean_force: np.ndarray,
    dx: float,
    equil_temp: float = 300.0,
    method: str = "trapezoid",
) -> tuple:
    """numeric integration of thermodynamic force by simpson, trapezoid or rectangle rule

    Args:
        data: data to integrate
        dx: bin width
        equil_temp: equilibrium temperature
        method: use 'simpson', 'trapezoid' or 'rectangle' rule

    Returns:
        pmf (np.ndarray): potential of mean force
        rho (np.ndarray): probability density
    """
    RT = R_in_SI * equil_temp / 1000.0
    data = np.copy(mean_force)

    if method == "simpson":
        data[1:-1] *= 2.0
        data[2:-1:2] *= 2.0
        pmf = np.full(len(data), dx / 3.0)
    elif method == "trapezoid":
        data[1:-1] *= 2.0
        pmf = np.full(len(data), dx / 2.0)
    else:  # method == 'rectangle':
        pmf = np.full(len(data), dx)

    for i in range(len(pmf)):
        pmf[i] *= data[0 : i + 1].sum()

    rho = np.exp(-pmf / RT)
    rho /= rho.sum() * dx

    pmf = -RT * np.log(rho, out=np.full_like(rho, np.nan), where=(rho != 0))
    return pmf, rho


def czar(
    grid: np.ndarray,
    xi: np.ndarray,
    la: np.ndarray,
    sigma: float,
    equil_temp: float = 300.0,
) -> np.ndarray:
    """Corrected z-averaged Restrained (CZAR)

       see: Lesage et. al., J. Phys. Chem. B (2017); https://doi.org/10.1021/acs.jpcb.6b10055

    Args:
        grid: grid for reaction coordinate
        xi: trajectory of reaction coordinate
        la: trajectory of fictitious particle
        sigma: thermal width of coupling of la to xi
        equil_temp: Temperature of the simulation

    Returns:
        mean_force: thermodynamic force (gradient of PMF)
    """
    RT = R_in_SI * equil_temp / 1000.0

    dx2 = (grid[1] - grid[0]) / 2.0
    grid_local = grid + dx2

    # get force constant from sigma
    k = RT / (sigma * sigma)

    hist = np.zeros(len(grid), dtype=float)
    f_corr = np.zeros(len(grid), dtype=float)
    for i, x in enumerate(grid_local):
        la_x = la[np.where(np.logical_and(xi >= x - dx2, xi < x + dx2))]

        hist[i] = len(la_x)
        if hist[i] > 0:
            la_avg = np.average(la_x)
            f_corr[i] = k * (la_avg - x)

    log_hist = np.log(hist, out=np.zeros_like(hist), where=(hist != 0))

    return -RT * np.gradient(log_hist, grid_local) + f_corr



import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 8):
    sys.exit("Sorry, Python < 3.8 is not supported")

setup(
    name="adaptive_sampling",
    packages=[
        "adaptive_sampling",
        "adaptive_sampling.colvars",
        "adaptive_sampling.processing_tools",
        "adaptive_sampling.sampling_tools"
    ],
    version="2.0.1",
    license="MIT",
    description="Sampling algorithms for molecular transitions",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Andreas Hulm",
    author_email="andreas.hulm@cup.uni-muenchen.de",
    url="https://github.com/ochsenfeld-lab/adaptive_sampling",
    download_url="https://github.com/ochsenfeld-lab/adaptive_sampling/archive/refs/tags/v2.0.1.zip",
    keywords=[
        "computational chemistry",
        "molecular dynamics",
        "free energy",
        "chemical reactions",
    ],
    install_requires=[
        "torch>=1.10.2",
        "numpy>=1.19.5",
        "scipy>=1.7.0",
    ],
    setup_requires=["pytest"],
    #test_requires=["pytest"],
    classifiers=[
        "Development Status :: 4 - Beta",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
    ],
    zip_safe=False,
)



#!/usr/bin/env python
import sys, os
import numpy as np
from typing import Tuple
from abc import ABC, abstractmethod

from ..interface.sampling_data import MDInterface
from ..colvars.colvars import CV
from .utils import diff
from ..units import *


class EnhancedSampling(ABC):
    """Abstract class for molecular dynamics based sampling algorithms

    Args:
        md: Object of the MD Inteface
        cv_def: definition of the Collective Variable (CV) (see adaptive_sampling.colvars)
            [["cv_type", [atom_indices], minimum, maximum, bin_width], [possible second dimension]]
        equil_temp: equillibrium temperature of MD
        verbose: print verbose information
        kinetice: calculate necessary data to obtain kinetics of reaction
        f_conf: force constant for confinement of system to the range of interest in CV space
        output_freq: frequency in steps for writing outputs
        multiple_walker: share bias with other simulations via buffer file
    """

    def __init__(
        self,
        md: MDInterface,
        cv_def: list,
        equil_temp: float = 300.0,
        verbose: bool = True,
        kinetics: bool = False,
        f_conf: float = 100,
        output_freq: int = 100,
        multiple_walker: bool = False,
        **kwargs,
    ):

        self.the_md = md
        self.the_cv = CV(self.the_md, requires_grad=True)
        self.out_freq = output_freq
        self.equil_temp = equil_temp
        self.verbose = verbose
        self.shared = multiple_walker

        # definition of CVs
        self.ncoords = len(cv_def)
        self.cv = np.array([item[0] for item in cv_def], dtype=str)
        self.atoms = [cv_def[i][1] for i in range(self.ncoords)]
        self.minx = np.array([item[2] for item in cv_def], dtype=float)
        self.maxx = np.array([item[3] for item in cv_def], dtype=float)
        self.dx = np.array([item[4] for item in cv_def], dtype=float)
        self.f_conf = np.array([f_conf for _ in range(self.ncoords)], dtype=float)

        self.cv_type = ["" for _ in range(self.ncoords)]
        (xi, delta_xi) = self.get_cv(**kwargs)

        # unit conversion
        self.minx, self.maxx, self.dx = self.unit_conversion_cv(
            self.minx, self.maxx, self.dx
        )
        self.f_conf = self.unit_conversion_force(self.f_conf)

        # store trajectories of CVs and temperature and epot between outputs
        md_state = self.the_md.get_sampling_data()
        self.traj = np.array([xi])
        self.temp = [md_state.temp]
        self.epot = [md_state.epot]

        # get number of bins (1D or 2D)
        self.nbins_per_dim = np.array([1, 1])
        self.grid = []
        for i in range(self.ncoords):
            self.nbins_per_dim[i] = (
                int(np.ceil(np.abs(self.maxx[i] - self.minx[i]) / self.dx[i]))
                if self.dx[i] > 0
                else 1
            )
            self.grid.append(
                np.linspace(
                    self.minx[i] + self.dx[i] / 2,
                    self.maxx[i] - self.dx[i] / 2,
                    self.nbins_per_dim[i],
                )
            )
        self.nbins = np.prod(self.nbins_per_dim)

        # accumulators and conditional averages
        self.bias = np.zeros(
            (self.ncoords, self.nbins_per_dim[1], self.nbins_per_dim[0]), dtype=float
        )
        self.cv_crit = np.copy(self.bias)

        self.histogram = np.zeros(
            (self.nbins_per_dim[1], self.nbins_per_dim[0]), dtype=float
        )
        self.pmf = np.copy(self.histogram)

        # trajectory of mass of CV for postprocessing of kinetics
        self.kinetics = kinetics
        if self.kinetics and self.ncoords == 1:
            self.mass_traj = [self._get_mass_of_cv(delta_xi)]
            self.abs_forces = [np.linalg.norm(md_state.forces)]
            self.CV_crit_traj = [np.abs(np.dot(md_state.forces, delta_xi[0]))]
            self.abs_grad_xi = [np.linalg.norm(delta_xi)]
        elif self.kinetics:
            self.kinetics = False
            if verbose:
                print(
                    " >>> Warning: kinetics only available for 1D collective variables"
                )

        if self.verbose:
            for i in range(self.ncoords):
                print(f"\n Initialize {self.cv[i]} as collective variable:")
                if self.cv_type[i] == "angle":
                    output_dat = (
                        self.minx[i] * DEGREES_per_RADIAN,
                        self.maxx[i] * DEGREES_per_RADIAN,
                        self.dx[i] * DEGREES_per_RADIAN,
                        "Degree",
                    )
                elif self.cv_type[i] == "distance":
                    output_dat = (
                        self.minx[i] * BOHR_to_ANGSTROM,
                        self.maxx[i] * BOHR_to_ANGSTROM,
                        self.dx[i] * BOHR_to_ANGSTROM,
                        "Angstrom",
                    )
                else:
                    output_dat = (self.minx[i], self.maxx[i], self.dx[i], "")
                print(f"\t Minimum{i}:\t{output_dat[0]} {output_dat[3]}")
                print(f"\t Maximum{i}:\t{output_dat[1]} {output_dat[3]}")
                print(f"\t Bin width{i}:\t{output_dat[2]} {output_dat[3]}")
            print("\t----------------------------------------------")
            print(f"\t Total number of bins:\t\t{self.nbins}\n")

    @abstractmethod
    def step_bias(self):
        pass

    @abstractmethod
    def get_pmf(self):
        pass

    @abstractmethod
    def shared_bias(self):
        pass

    @abstractmethod
    def write_restart(self):
        pass

    @abstractmethod
    def restart(self):
        pass

    @abstractmethod
    def write_traj(self):
        pass

    def harmonic_walls(
        self,
        xi: np.ndarray,
        delta_xi: np.ndarray,
        margin: np.ndarray = np.array([0, 0]),
    ) -> np.ndarray:
        """confine system with harmonic walls to range(self.minx, self.maxx)

        Args:
            xi: collective variable
            delta_xi: gradient of collective variable
            margin: inset for start of harmonic wall

        Returns:
            bias_force: confinement force
        """
        conf_force = np.zeros_like(self.the_md.get_sampling_data().forces.ravel())

        for i in range(self.ncoords):
            if xi[i] > (self.maxx[i] - margin[i]):
                r = diff(self.maxx[i] - margin[i], xi[i], self.cv_type[i])
                conf_force -= self.f_conf[i] * r * delta_xi[i]

            elif xi[i] < (self.minx[i] + margin[i]):
                r = diff(self.minx[i] + margin[i], xi[i], self.cv_type[i])
                conf_force -= self.f_conf[i] * r * delta_xi[i]

        return conf_force

    def get_index(self, xi: np.ndarray) -> list:
        """get list of bin indices for current position of CVs or extended variables

        Args:
            xi (np.ndarray): Current value of collective variable

        Returns:
            bin_x (list):
        """
        bin_x = [-1, -1]
        for i in range(self.ncoords):
            bin_x[i] = int(np.floor(np.abs(xi[i] - self.minx[i]) / self.dx[i]))
        return bin_x

    def _check_boundaries(self, x):
        """returns True if x is between minx and maxx"""
        return (x < self.maxx).all() and (x > self.minx).all()

    def unit_conversion_cv(self, *args):
        """convert input to atomic units

        Args:
            args: arrays to convert of size(dimensions)

        Returns:
            args in atomic units
        """
        for i in range(self.ncoords):
            for arg in args:
                if self.the_cv.type == "angle":
                    arg[i] /= DEGREES_per_RADIAN
                elif self.the_cv.type == "distance":
                    arg[i] /= BOHR_to_ANGSTROM
        return args

    def unit_conversion_force(self, *args):
        """convert input to atomic units

        Args:
            *args: arrays to convert of size(dimensions)

        Returns:
            args in atomic units
        """

        for i in range(self.ncoords):
            for arg in args:
                if self.cv_type == "angle":
                    arg[i] *= DEGREES_per_RADIAN * DEGREES_per_RADIAN / atomic_to_kJmol
                elif self.cv_type == "distance":
                    arg[i] *= BOHR_to_ANGSTROM * BOHR_to_ANGSTROM / atomic_to_kJmol
                else:
                    arg[i] /= atomic_to_kJmol
        return args

    def get_cv(self, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """get state of collective variable

        Returns:
            xi: state of the collective variable
            grad_xi: gradient of the collective variable
        """
        self.the_cv.requires_grad = True
        xi = np.zeros(self.ncoords)
        grad_xi = np.zeros(
            (self.ncoords, len(self.the_md.get_sampling_data().forces.ravel()))
        )

        for i in range(self.ncoords):
            xi[i], grad_xi[i] = self.the_cv.get_cv(self.cv[i], self.atoms[i], **kwargs)
            self.cv_type[i] = self.the_cv.type

        return xi, grad_xi

    def _kinetics(self, delta_xi):
        """accumulates data for kinetics"""
        forces = self.the_md.get_sampling_data().forces
        self.mass_traj.append(self._get_mass_of_cv(delta_xi))
        self.abs_forces.append(np.linalg.norm(forces))
        self.CV_crit_traj.append(np.dot(delta_xi[0], forces))
        self.abs_grad_xi.append(np.linalg.norm(delta_xi))

    def _get_mass_of_cv(self, delta_xi: np.ndarray) -> float:
        """get mass of collective variable for TS theory and kinetics
        only defined for 1D reaction coordinates!

        Args:
            delta_xi: gradients of cv's

        Returns:
            m_xi_inv: coordinate dependent mass of collective variabl
        """
        if self.ncoords == 1:
            return np.dot(
                delta_xi[0], (1.0 / np.repeat(self.the_md.mass, 3)) * delta_xi[0]
            )
        else:
            return 0.0

    def write_output(self, data: dict, filename="free_energy.dat"):
        """write results to output file

        Args:
            data: results to write
            filename: name of output file
        """
        grid = np.copy(self.grid)
        for i in range(self.ncoords):
            if self.the_cv.type == "angle":
                grid *= DEGREES_per_RADIAN
            elif self.the_cv.type == "distance":
                grid *= BOHR_to_ANGSTROM

        # head of data columns
        with open(filename, "w") as fout:
            for i in range(self.ncoords):
                fout.write("%14s\t" % "CV{dim}".format(dim=i))
            for kw in data.keys():
                fout.write("%14s\t" % kw)
            fout.write("\n")

            # write data to columns
            if self.ncoords == 1:
                for i in range(self.nbins):
                    fout.write("%14.6f\t" % grid[0][i])
                    for dat in data.values():
                        fout.write("%14.6f\t" % dat[0][i])
                    fout.write("\n")

            if self.ncoords == 2:
                for i in range(self.nbins_per_dim[1]):
                    for j in range(self.nbins_per_dim[0]):
                        fout.write("%14.6f\t%14.6f\t" % (grid[1][i], grid[0][j]))
                        for dat in data.values():
                            fout.write("%14.6f\t" % dat[i, j])
                        fout.write("\n")

    def _write_traj(self, data: dict = {}, filename: str = 'CV_traj.dat'):
        """write trajectory of extended or normal ABF at output times

        Args:
            extended: True for use extended system
            additional_data: additional data that one might want to write
        """
        step = self.the_md.get_sampling_data().step

        # convert units to degree and Angstrom
        for i in range(self.ncoords):
            if self.cv_type[i] == "angle":
                self.traj[:, i] *= DEGREES_per_RADIAN
            elif self.cv_type[i] == "distance":
                self.traj[:, i] *= BOHR_to_ANGSTROM

        # write header
        if not os.path.isfile(filename) and step == 0:
            # start new file in first step

            with open(filename, "w") as traj_out:
                traj_out.write("%14s\t" % "time [fs]")
                for i in range(len(self.traj[0])):
                    traj_out.write("%14s\t" % f"Xi{i}")
                for kw in data.keys():
                    traj_out.write("%14s\t" % kw)
                if self.kinetics:
                    traj_out.write("%14s\t" % "m_xi_inv [a.u.]")
                    traj_out.write("%14s\t" % "|dU| [a.u.]")
                    traj_out.write("%14s\t" % "|dxi| [a.u.]")
                    traj_out.write("%14s\t" % "dU*dxi [a.u.]")

        elif step > 0:
            # append new steps of trajectory since last output
            with open(filename, "a") as traj_out:
                for n in range(self.out_freq):
                    traj_out.write(
                        "\n%14.6f\t"
                        % (
                            (step - self.out_freq + n)
                            * self.the_md.get_sampling_data().dt
                        )
                    )  # time in fs
                    for i in range(len(self.traj[0])):
                        traj_out.write("%14.6f\t" % (self.traj[-self.out_freq + n][i]))
                    for val in data.values():
                        traj_out.write("%14.6f\t" % (val[-self.out_freq + n]))
                    if self.kinetics:
                        traj_out.write(
                            "%14.6f\t" % (self.mass_traj[-self.out_freq + n])
                        )
                        traj_out.write(
                            "%14.6f\t" % (self.abs_forces[-self.out_freq + n])
                        )
                        traj_out.write(
                            "%14.6f\t" % (self.abs_grad_xi[-self.out_freq + n])
                        )
                        traj_out.write(
                            "%14.6f\t" % (self.CV_crit_traj[-self.out_freq + n])
                        )

    def _write_restart(self, filename, **kwargs):
        """save **kwargs in .npz file"""
        np.savez(filename, **kwargs)
        sys.stdout.flush()



import numpy as np
from .enhanced_sampling import EnhancedSampling
from .metaeabf import WTMeABF
from .gamd import GaMD
from .utils import welford_var, diff, cond_avg
from ..processing_tools.thermodynamic_integration import integrate
from ..units import *


class GaWTMeABF(WTMeABF, GaMD, EnhancedSampling):
    """Gaussian-accelerated Well-Tempered Metadynamics extended-system Adaptive Biasing Force Method

    see: Chen et. al., J. Chem. Theory Comput. (2021); https://doi.org/10.1021/acs.jctc.1c00103

    The collective variable (CV) is coupled to a fictitious particle with an harmonic force.
    The dynamics of the fictitious particel is biased using a combination of ABF and metadynamics.
    The dynamics of the pysical system is biased with CV-independend an GaMD boost potential

    Args:
        ext_sigma: thermal width of coupling between collective and extended variable
        ext_mass: mass of extended variable in atomic units

        nfull: Number of force samples per bin where full bias is applied,
               if nsamples < nfull the bias force is scaled down by nsamples/nfull
        friction: friction coefficient for Lagevin dynamics of the extended-system
        seed_in: random seed for Langevin dynamics of extended-system
        hill_height: height of Gaussian hills in kJ/mol
        hill_std: standard deviation of Gaussian hills in units of the CV (can be Bohr, Degree, or None)
        gamd_sigma0: upper limit of standard deviation of boost potential
        gamd_init_step: initial steps where no bias is applied to estimate min, max and var of potential energy
        gamd_equil_steps: equilibration steps, min, max and var of potential energy is still updated
                          force constant of coupling is calculated from previous steps
        md: Object of the MD Inteface
        cv_def: definition of the Collective Variable (CV) (see adaptive_sampling.colvars)
                [["cv_type", [atom_indices], minimum, maximum, bin_width], [possible second dimension]]
        do_wtm: if False, no metadynamics potential is applied (Gaussian-accelerated eABF)
        hill_drop_freq: frequency of hill creation in steps
        well_tempered_temp: effective temperature for WTM, if None, hills are not scaled down (normal metadynamics)
        force_from_grid: forces are accumulated on grid for performance,
                         if False, forces are calculated from sum of Gaussians in every step
        gamd_bound: "lower": use lower bound for GaMD boost
                    "upper: use upper bound for GaMD boost
        equil_temp: equillibrium temperature of MD
        verbose: print verbose information
        kinetice: calculate necessary data to obtain kinetics of reaction
        f_conf: force constant for confinement of system to the range of interest in CV space
        output_freq: frequency in steps for writing output
    """

    def __init__(self, *args, qm_boost: bool = False, do_wtm: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_wtm = do_wtm
        self.qm_boost = qm_boost

    def step_bias(self, write_output: bool = True, write_traj: bool = True, **kwargs):

        md_state = self.the_md.get_sampling_data()
        epot = md_state.epot
        if self.qm_boost:
            epot = md_state.epot - md_state.mm_epot
            self.gamd_forces = np.copy(md_state.qm_force)
        else:
            epot = md_state.epot
            self.gamd_forces = np.copy(md_state.forces)

        (xi, delta_xi) = self.get_cv(**kwargs)

        self._propagate()
        bias_force = self._extended_dynamics(xi, delta_xi)

        if md_state.step < self.gamd_init_steps:
            self._update_pot_distribution(epot)

        else:
            if md_state.step == self.gamd_init_steps:
                self._calc_E_k0()

            # apply gamd boost potential
            prefac = self.k0 / (self.pot_max - self.pot_min)
            self.gamd_pot = 0.5 * prefac * np.power(self.E - epot, 2)
            bias_force -= prefac * (self.E - epot) * self.gamd_forces

            if md_state.step < self.gamd_equil_steps:
                self._update_pot_distribution(epot)
                self._calc_E_k0()

            else:
                # (WTM-)eABF bias on extended-variable only in production
                if self.do_wtm:
                    mtd_forces = self.get_wtm_force(self.ext_coords)

                if self._check_boundaries(self.ext_coords):

                    bink = self.get_index(self.ext_coords)
                    self.ext_hist[bink[1], bink[0]] += 1

                    for i in range(self.ncoords):

                        # linear ramp function
                        ramp = (
                            1.0
                            if self.ext_hist[bink[1], bink[0]] > self.nfull
                            else self.ext_hist[bink[1], bink[0]] / self.nfull
                        )

                        # apply (WTM-)eABF bias force on extended variable
                        (
                            self.abf_forces[i][bink[1], bink[0]],
                            self.m2_force[i][bink[1], bink[0]],
                            self.var_force[i][bink[1], bink[0]],
                        ) = welford_var(
                            self.ext_hist[bink[1], bink[0]],
                            self.abf_forces[i][bink[1], bink[0]],
                            self.m2_force[i][bink[1], bink[0]],
                            self.ext_k[i]
                            * diff(self.ext_coords[i], xi[i], self.cv_type[i]),
                        )
                        self.ext_forces -= ramp * self.abf_forces[i][bink[1], bink[0]]

                        if self.do_wtm:
                            self.ext_forces += mtd_forces[i]

        # free energy reweighting
        if self._check_boundaries(xi):

            bink = self.get_index(xi)
            self.histogram[bink[1], bink[0]] += 1

            # CZAR
            for i in range(self.ncoords):
                dx = diff(self.ext_coords[i], self.grid[i][bink[i]], self.cv_type[i])
                self.correction_czar[i][bink[1], bink[0]] += self.ext_k[i] * dx

            # GaMD
            if md_state.step >= self.gamd_equil_steps:

                (
                    self.gamd_c1[bink[1], bink[0]],
                    self.gamd_m2[bink[1], bink[0]],
                    self.gamd_c2[bink[1], bink[0]],
                ) = welford_var(
                    self.histogram[bink[1], bink[0]],
                    self.gamd_c1[bink[1], bink[0]],
                    self.gamd_m2[bink[1], bink[0]],
                    self.gamd_pot,
                )

        self._up_momenta()

        self.traj = np.append(self.traj, [xi], axis=0)
        self.ext_traj = np.append(self.ext_traj, [self.ext_coords], axis=0)
        self.temp.append(md_state.temp)
        self.gamd_pot_traj.append(self.gamd_pot)
        self.epot.append(md_state.epot)

        # correction for kinetics
        if self.kinetics:
            self._kinetics(delta_xi)

        # write traj and output
        if md_state.step % self.out_freq == 0:

            if write_traj:
                self.write_traj()

            if write_output:
                self.get_pmf()
                output = {"hist": self.histogram, "free energy": self.pmf}
                for i in range(self.ncoords):
                    output[f"abf force {i}"] = self.abf_forces[i]
                    output[f"czar force {i}"] = self.czar_force[i]
                output[f"metapot"] = self.metapot
                output[f"GaMD corr"] = self.gamd_corr

                self.write_output(output, filename="gaeabf.out")
                self.write_restart()

        return bias_force

    def get_pmf(self, method: str = "trapezoid"):

        log_rho = np.log(
            self.histogram,
            out=np.zeros_like(self.histogram),
            where=(0 != self.histogram),
        )
        avg_force = cond_avg(self.correction_czar, self.histogram)

        self.gamd_corr = -self.gamd_c1 - self.gamd_c2 / (
            2.0 * kB_in_atomic * self.equil_temp
        )

        if self.ncoords == 1:
            self.czar_force[0] = (
                -kB_in_atomic * self.equil_temp * np.gradient(log_rho[0], self.grid[0])
                + avg_force[0]
            )
            self.pmf[0, :], _ = integrate(
                self.czar_force[0][0],
                self.dx,
                equil_temp=self.equil_temp,
                method=method,
            )
            self.pmf += self.gamd_corr
            self.pmf *= atomic_to_kJmol
            self.pmf -= self.pmf.min()

        else:
            der_log_rho = np.gradient(log_rho, self.grid[1], self.grid[0])
            self.czar_force[0] = (
                -kB_in_atomic * self.equil_temp * der_log_rho[1] + avg_force[0]
            )
            self.czar_force[1] = (
                -kB_in_atomic * self.equil_temp * der_log_rho[0] + avg_force[1]
            )
            if self.verbose:
                print(
                    " >>> Info: On-the-fly integration only available for 1D coordinates"
                )

    def shared_bias(self):
        """TODO"""
        pass

    def write_restart(self, filename: str = "restart_gaabf"):
        """write restart file

        Args:
            filename: name of restart file
        """
        self._write_restart(
            filename=filename,
            hist=self.histogram,
            force=self.bias,
            var=self.var_force,
            m2=self.m2_force,
            ext_hist=self.ext_hist,
            czar_corr=self.correction_czar,
            abf_force=self.abf_forces,
            center=self.center,
            metapot=self.metapot,
            gamd_c1=self.gamd_c1,
            gamd_m2=self.gamd_m2,
            corr=self.gamd_corr,
            pot_count=self.pot_count,
            pot_var=self.pot_var,
            pot_std=self.pot_std,
            pot_m2=self.pot_m2,
            pot_avg=self.pot_avg,
            pot_min=self.pot_min,
            pot_max=self.pot_max,
            k0=self.k0,
        )

    def restart(self, filename: str = "restart_gaabf"):
        """restart from restart file

        Args:
            filename: name of restart file
        """
        try:
            data = np.load(filename + ".npz", allow_pickle=True)
        except:
            raise OSError(f" >>> fatal error: restart file {filename}.npz not found!")

        self.histogram = data["hist"]
        self.bias = data["force"]
        self.var_force = data["var"]
        self.m2_force = data["m2"]
        self.ext_hist = data["ext_hist"]
        self.correction_czar = data["czar_corr"]
        self.abf_forces = data["abf_force"]
        self.center = data["center"].tolist()
        self.metapot = data["self.metapot"]
        self.gamd_c1 = data["c1"]
        self.gamd_m2 = data["m2"]
        self.gamd_corr = data["corr"]
        self.pot_count = data["pot_count"]
        self.pot_var = data["pot_var"]
        self.pot_std = data["pot_std"]
        self.pot_m2 = data["pot_m2"]
        self.pot_avg = data["pot_avg"]
        self.pot_min = data["pot_min"]
        self.pot_max = data["pot_max"]
        self.k0 = data["k0"]

        if self.verbose:
            print(f" >>> Info: Adaptive sampling restartet from {filename}!")

    def write_traj(self):
        """save trajectory for post-processing"""

        data = self._write_ext_traj()
        data[f"E_gamd [H]"] = self.gamd_pot_traj
        data["Epot [H]"] = self.epot
        data["T [K]"] = self.temp

        self._write_traj(data)

        # reset trajectories to save memory
        self.traj = np.array([self.traj[-1]])
        self.ext_traj = np.array([self.ext_traj[-1]])
        self.epot = [self.epot[-1]]
        self.temp = [self.temp[-1]]
        self.gamd_pot_traj = [self.gamd_pot_traj[-1]]
